import numpy as np
from scipy.interpolate import griddata
from scipy.optimize import fsolve
import matplotlib.pyplot as plt

from pyXSteam.XSteam import XSteam
steamTable = XSteam(XSteam.UNIT_SYSTEM_MKS)  # m/kg/sec/°C/bar/W

import sys
import os

# CaCl2 property functions
# DONE
def saturation_temperature(x, p):
    """
    last change: brought to python by Dorian Höffner
    author:
    source:
    """
    T_guess = 20
    T = fsolve(lambda T: saturation_pressure(x, T) - p, T_guess)[0]

    # check if the solution is in liquid state
    t_sol = solubility_temperature(x)[0]
    if T < t_sol:
        print(f"Warning: The temperature {T} is below the solubility temperature {t_sol} at x={x}.")
    
    return T

# TODO
def enthalpy(x, T):
    """
    Last Change: brought to python by Dorian Höffner
    Autor:
    Quelle:

    Parameter:
     T in [°C]
     x = m_NaOH/(m_H2O + m_NaOH)
     h in [kJ/kg]
     """
    
    # sensible: use water as reference

    # excess enthalpy based on dilution ???

    print("Enthalpy function not available yet. Need to use differential_enthalpy_AD instead.")

    # check if the solution is in liquid state
    t_sol = solubility_temperature(x)[0]
    if T < t_sol:
        print(f"Warning: The temperature {T} is below the solubility temperature {t_sol} at x={x}.")

    return # enthalpy in [kJ/kg]

def differential_enthalpy_AD(x, T):
    """
    Calculates the differential enthalpy [kJ/kgH2O] of a CaCl2 solution based on concentration and temperature.
    How much additional energy (in comparison to pure vaporization enthalpy) is needed to vaporize a certain amount of water in a CaCl2 solution.
    
    Parameters:
        x (float): Concentration of CaCl2 in the solution [kg/kg].
        T (float): Temperature of the solution [°C].

    Returns:
        float: Differential enthalpy [kJ/kgH2O].

    Author: Dorian Höffner 11/2024
    Source: CONDE2009 "Aqueous solutions of lithium and calcium chlorides: property formulations for use in air conditioning equipment design" (2009)
    """

    H1 = 0.855
    H2 = -1.965
    H3 = -2.265
    H4 = 0.8
    H5 = -955.690
    H6 = 3011.974

    # Reduced Temperature (with critical T of water)
    Theta = (T + 273.15) / 647.1 # FLAG

    # zeta
    zeta = x / (H4-x)

    # Reference differental enthalpy
    dh_dil0 =  H5 + H6 * Theta

    # differental enthalpy
    dh_dil = dh_dil0 * (1 + (zeta/H1)**H2)**H3

    # check if the solution is in liquid state
    t_sol = solubility_temperature(x)[0]
    if T < t_sol:
        print(f"Warning: The temperature {T} is below the solubility temperature {t_sol} at x={x}.")

    return dh_dil

# DONE
def saturation_pressure(x, T):
    """
    Calculates the pressure of aqueous NaOH-H2O solutions at high concentrations.
    
    Last Change: Dorian Höffner 10/2024
    Author: O. Buchin, 03/2011
    Source: Correlation according to CONDE2009 "Aqueous solutions of lithium and calcium chlorides:
    property formulations for use in air conditioning equipment design" (2009)
    
    Parameters:
    T (array-like): Temperature in [°C].
    x (array-like): Mass fraction, defined as m_NaOH / (m_H2O + m_NaOH).
    
    Returns:
    pd (float or array-like): Pressure in [Pa]
    """
    zeta = x  # Mass fraction of salt in solution = Msalt / Msolution

    # Coefficients
    pi0 = 0.31
    pi1 = 3.698
    pi2 = 0.6
    pi3 = 0.231
    pi4 = 4.584
    pi5 = 0.49
    pi6 = 0.478
    pi7 = -5.2
    pi8 = -0.4
    pi9 = 0.018

    A = 2 - (1 + (zeta / pi0) ** pi1) ** pi2
    B = (1 + (zeta / pi3) ** pi4) ** pi5 - 1
    pi25 = (
        1
        - (1 + (zeta / pi6) ** pi7) ** pi8
        - pi9 * np.exp(-((zeta - 0.1) ** 2) / 0.005)
    )

    # Theta - reduced temperature
    TcH2O = 647.26  # K
    pcH2O = 22.064  # MPa
    Theta = (T + 273.15) / TcH2O  # Corrected variable name from 't' to 'T'

    fsol = A + B * Theta
    pi = pi25 * fsol

    # Vapor pressure of water
    tau = 1 - Theta

    A0 = -7.858230
    A1 = 1.839910
    A2 = -11.781100
    A3 = 22.670500
    A4 = -15.939300
    A5 = 1.775160

    lnpi = (
        A0 * tau
        + A1 * tau ** 1.5
        + A2 * tau ** 3
        + A3 * tau ** 3.5
        + A4 * tau ** 4
        + A5 * tau ** 7.5
    ) / (1 - tau)
    pdH2O = np.exp(lnpi) * pcH2O * 1e6  # Convert from MPa to Pa

    pd = pi * pdH2O

    # check if the solution is in liquid state
    t_sol = solubility_temperature(x)[0]
    if T < t_sol:
        print(f"Warning: The temperature {T} is below the solubility temperature {t_sol} at x={x}.")

    return pd

# DONE
def saturation_concentration(p, T):
    """
    Calculates the saturation concentration of NaOH in water based on the temperature and pressure.

    Author: Dorian Höffner
    Date: 2024-09-17

    Parameters:
        p (float): Pressure in Pa.
        T (float): Temperature in °C.

    Returns:
        float: Saturation concentration in kg NaOH / kg solution.
    """

    # Calculate the saturation concentration
    x = 0.001
    while saturation_pressure(x, T) - p > 1:   # 1 Pa tolerance
        x += 0.001

    # check if the solution is in liquid state
    t_sol = solubility_temperature(x)[0]
    if T < t_sol:
        print(f"Warning: The temperature {T} is below the solubility temperature {t_sol} at x={x}.")

    return x

# DONE
def density(x, T):
    """
    Calculates the density of CaCl2 solution based on concentration and temperature.

    Parameters:
    x (float): Mole fraction of NaOH in the solution. [m_CaCl2 / (m_h2o+m_CaCl2)]
    T (float): Temperature in °C.

    Returns:
    float: Density in kg/m^3.
    ---
    Author: Dorian Höffner 11/2024
    Source: CONDE2009 "Aqueous solutions of lithium and calcium chlorides: property formulations for use in air conditioning equipment design" (2009)
    """
    if x > 0.6:
        raise ValueError("Concentration of CaCl2 cannot be greater than 0.6 kg/kg. Correlation is not valid for this concentration.")
    
    # check if the solution is in liquid state
    t_sol = solubility_temperature(x)[0]
    if T < t_sol:
        print(f"Warning: The temperature {T} is below the solubility temperature {t_sol} at x={x}.")

    # Water Correlation
    def rho_h2o(T):
        B = [1.9937718430, 1.0985211604, -0.5094492996, -1.7619124270, -44.9005480267, -723692.2618632]
        rho_h2o_c = 322.0  # kg/m^3 (critical density of water)
        Tc = 647.1  # K
        Theta = (T+273.15) / Tc
        t = 1-Theta
        rho_h2o = rho_h2o_c * (1 + B[0] * t**(1/3) + B[1] * t**(2/3) + B[2] * t**5/3 + B[3] * t**(16/3) + B[4] * t**(43/3) + B[5] * t**(110/3))
        return rho_h2o
    
    # CaCl2-H2O Correlation
    c = [1.0, 0.836014, -0.436300, 0.105642]
    rho = rho_h2o(T) * sum([c[i] * (x/(1-x))**i for i in range(4)])

    return rho # density in kg/m^3

# # DONE
# def density_old(x, T):
#     """
#     Calculates the density of CaCl2 solution based on concentration and temperature.

#     Parameters:
#     x (float): Mole fraction of CaCl2 in the solution. [m_CaCl2 / (m_h2o+m_CaCl2)]
#     T (float): Temperature in °C.

#     Returns:
#     float: Density in kg/m^3.
#     ---
#     Author: Dorian Höffner 11/2024
#     Data extracted from: Hao et al. 2016 "Modeling of CO2 solubility in single and mixed electrolyte solutions using statistical associating fluid theory"
#     """
#     # Convert molality to concentration in kg/kg for CaCl solution
#     molalities = np.array([
#         [0.2380952380952381, 0.7471264367816093, 1.2807881773399015, 1.781609195402299, 2.3070607553366176,
#         2.889983579638752, 3.472906403940887, 4.010673234811166, 4.486863711001642, 4.909688013136289],
#         [0.24630541871921183, 0.7594417077175698, 1.293103448275862, 1.7980295566502464, 2.323481116584565,
#         2.9105090311986865, 3.477011494252874, 4.018883415435139, 4.503284072249589, 4.909688013136289],
#         [0.2545155993431856, 0.7881773399014779, 1.3382594417077176, 1.8267651888341545, 2.3399014778325125,
#         2.935139573070608, 3.493431855500821, 4.0353037766830875, 4.5073891625615765, 4.917898193760263],
#         [0.2504105090311987, 0.8087027914614122, 1.3669950738916257, 1.8596059113300494, 2.376847290640394,
#         2.9761904761904763, 3.501642036124795, 4.059934318555008, 4.527914614121511, 4.946633825944171],
#         [0.513136288998358, 0.8292282430213465, 1.416256157635468, 1.8308702791461413, 2.389162561576355,
#         3.0049261083743843, 3.5180623973727423, 4.072249589490969, 4.499178981937603, 4.9384236453201975]
#     ])

#     M_CaCl2 = 110.98  # g/mol for CaCl2
#     x_values = molalities * M_CaCl2 / 1000  # Convert molality to kg/kg

#     temperatures = [
#         [298.15] * 10,
#         [323.15] * 10,
#         [373.15] * 10,
#         [423.15] * 10,
#         [473.15] * 10
#     ]

#     densities = [
#         [1.026256077795786, 1.0656401944894651, 1.1055105348460292, 1.1434359805510534, 1.1794165316045382,
#         1.220259319286872, 1.2572123176661265, 1.2922204213938413, 1.3218800648298217, 1.347163695299838],
#         [1.013128038897893, 1.0544570502431119, 1.0952998379254457, 1.1322528363047002, 1.1696920583468395,
#         1.2100486223662885, 1.247001620745543, 1.280551053484603, 1.3097244732576985, 1.3340356564019449],
#         [0.9810372771474878, 1.0257698541329012, 1.0675850891410048, 1.1055105348460292, 1.1414910858995138,
#         1.1828200972447327, 1.220745542949757, 1.2552674230145868, 1.2829821717990275, 1.3082658022690439],
#         [0.9397082658022691, 0.986871961102107, 1.033063209076175, 1.0709886547811993, 1.1094003241491086,
#         1.152674230145867, 1.1881685575364669, 1.2246353322528363, 1.253808752025932, 1.2781199351701784],
#         [0.9149108589951378, 0.9440842787682334, 0.993679092382496, 1.0286871961102106, 1.072933549432739,
#         1.1176661264181524, 1.1541329011345218, 1.1915721231766614, 1.2188006482982172, 1.2455429497568882]
#     ]

#     # Flatten data for interpolation
#     flat_concentrations = np.concatenate(x_values)
#     flat_temperatures = np.concatenate(temperatures)
#     flat_densities = np.concatenate(densities)

#     # Create grid data
#     points = np.array(list(zip(flat_concentrations, flat_temperatures)))
#     values = flat_densities

#     # Interpolate density at the given (x, T) point
#     return griddata(points, values, (x, T + 273.15), method='linear')
    

# NOT AVAILABLE YET
# DONE
def specific_heat_capacity(x, T):
    """
    Calculate the specific heat capacity of a CaCl2 solution (CaCl2-H2O) as a function of temperature and concentration.
    
    ---
    x: float
        Concentration of CaCl2 in the solution [kg/kg]
    T: float
        Temperature of the solution [°C]
    ---
    returns: float
        Specific heat capacity of the solution [kJ/kgK]
    ---
    Author: Dorian Höffner
    Date: 2024-11-04
    Source: Conde 2009 "Aqueous solutions of lithium and calcium chlorides: property formulations for use in air conditioning equipment design" (2009)
    """

    # Coefficients
    a = np.where(T <= 0, 830.54602, 88.7891)
    b = np.where(T <= 0, -1247.52013, -120.1958)
    c = np.where(T <= 0, -68.60350, -16.9264)
    d = np.where(T <= 0, 491.27650, 52.4654)
    e = np.where(T <= 0, -1.80692, 0.10826)
    f = np.where(T <= 0, -137.51511, 0.46988)

    A = 1.63799
    B = -1.69002
    C = 1.05124
    D = 0.0
    E = 0.0
    F = 58.5225
    G = -105.6343
    H = 47.7948

    # Theta - reduced temperature
    Theta = (T + 273.15) / 228 - 1

    # cpH2O calculation
    cpH2O = (a + b * Theta**0.02 + c * Theta**0.04 + 
             d * Theta**0.06 + e * Theta**1.8 + f * Theta**8)

    # f1 calculation
    f1 = A * x + B * x**2 + C * x**3

    # f2 calculation
    f2 = F * Theta**0.02 + G * Theta**0.04 + H * Theta**0.06

    # cp calculation
    cp = cpH2O * (1 - f1 * f2)

    return cp

    #return cp # specific heat capacity in kJ/kgK

# DONE
def dynamic_viscosity(x, T):
    """
    Calculate the dynamic viscosity of a NaOH solution as a function of temperature and concentration.
    The function is based on the following publication: "Aqueous solutions of lithium and calcium chlorides: property formulations for use in air conditioning equipment design" (2009)
    ---
    x: float
        Concentration of NaOH in the solution [kg/kg]
    T: float
        Temperature of the solution [°C]
    ---
    returns: float
        Dynamic viscosity of the solution [Pa s]
    ---
    Last Change: Dorian Höffner 10/2024
    Author: O. Buchin 09/2010
    """

    p = 1.01325  # bar

    # Coefficients
    eta1 = -0.169310
    eta2 = 0.817350
    eta3 = 0.574230
    eta4 = 0.398750
    
    # Intermediate values
    xi = x
    zeta = xi / ((1 - xi) ** (1 / 0.6))
    
    # Coefficients for eta_H2O calculation
    A = 1.0261862
    B = 12481.702
    C = -19510.923
    D = 7065.286
    E = -395.561
    F = 143922.996
    
    # Reduced temperature
    Theta = (T + 273.15) / 228 - 1

    # Calculate eta_H2O for T <= 0 and T > 0
    etaH2O_0 = steamTable.my_pt(p, 0.0000001)
    eta_H2O = np.where(T <= 0, 
                       etaH2O_0 * (A + B * Theta**0.02 + C * Theta**0.04 + D * Theta**0.08 + E * Theta**2.85 + F * Theta**8),
                       steamTable.my_pt(p, T))

    
    # Final viscosity calculation
    TcH2O = 647.26  # K
    Theta = (T + 273.15) / TcH2O
    eta = eta_H2O * np.exp(eta1 * zeta**3.6 + eta2 * zeta + eta3 * zeta / Theta + eta4 * zeta**2)
    
    # Output in mPas
    return float(eta * 1000)

# DONE
def diffusion_coefficient(x, T):
    """
    Computes the self diffusion coefficient of a CaCl2 solution.

    Parameters:
        T (np.ndarray or float): Temperature of the solution (in Celsius) (column vector).
        x (np.ndarray or float): Concentration of solution (0 to 1) kgSalt/kgSolution (column vector).
    
    Returns:
        float or np.ndarray: Self diffusion coefficient (m^2/s).

    ---
    Author: O. Buchin 03/2011 (original matlab)
    Last change: Dorian Höffner 11/2024
    Source: Holz, Manfred; Heil, Stefan R.; Sacco, Antonio "Temperature-dependent self-diffusion coefficients of water and six selected molecular liquids for calibration in accurate 1H NMR PFG measurements"
    """
    
    # Self Diffusion coefficient for water
    D0 = 1.635e-8  # m^2/s
    TS = 215.05    # K
    gamma = 2.063
    
    Dw = D0 * (((T + 273.15) / TS) - 1) ** gamma
    
    # Self Diffusion coefficient for solution
    d1 = 0.55
    d2 = -5.52
    d3 = -0.56
    
    D = Dw * (1 - (1 + (np.sqrt(x) / d1) ** d2) ** d3)
    
    # Return as float if input is scalar
    return float(D) if np.isscalar(D) else D

# NOT AVAILABLE YET
def thermal_conductivity(x, T, p):
    """
    Calculate the thermal conductivity of a NaOH solution as a function of temperature and concentration.
    The function is based on the following publication:
    ---
    x: float
        Concentration of NaOH in the solution [kg/kg]
    T: float
        Temperature of the solution [°C]
    ---
    returns: float
        Thermal conductivity of the solution [W/mK]
    ---
    """
    print("Thermal conductivity function not implemented yet.")

    #return lambda_NaOH # [W/mK]

# DONE
def pTDiagram(log=True, invT=True, editablePlot=False, show_percentages=True):
    """
    Plots the pressure-temperature diagram for NaOH-H2O solutions.
    Author: Dorian Höffner 2024-04-26
    Last Change: 2024-09-17
    
    Parameters:
        log (bool): If True, the y-axis will be logarithmic.
        invT (bool): If True, the x-axis will be scaled as -1/T.
        editablePlot (bool): If True, the plot will be editable.
        show_percentages (bool): If True, the concentrations will be labeled.

    Returns:
        None
    """

    # turn off print statements
    devnull = open(os.devnull, 'w')
    stdout_old = sys.stdout
    sys.stdout = devnull

    
    steamTable = XSteam(XSteam.UNIT_SYSTEM_MKS)  # m/kg/sec/°C/bar/W

    # get data to plot crystallization curve
    cryst_data = crystallization_curve(return_data=True)
    cryst_T = np.array([T for x, T in cryst_data]).flatten()
    cryst_p = np.array([saturation_pressure(x, T) for x, T in cryst_data]).flatten()
    
    # Temperature range
    temperaturesC = np.arange(0, 110, 1)
    temperaturesK = temperaturesC + 273.15
    concentrations = np.arange(0.1, 0.69, 0.01)
    cr_temperaturesC = cryst_T
    cr_temperaturesK = cryst_T + 273.15
    
    # Prepare the plot
    plt.figure(dpi=300)
    
    # these temperatures are used for the x-axis
    plotTemperatures = np.arange(0, 101, 10) + 273.15
    
    # Calculate water vapor pressure using XSteam
    waterPressure = [steamTable.psat_t(T - 273.15) * 1e5 for T in temperaturesK]  # convert bar to Pa
    
    # Plot the data
    for x in concentrations:

        # Calculate saturation pressure for each temperature at the given concentration
        p = []
        for T in temperaturesC:
            p.append(saturation_pressure(x, T))  # Assuming PressureNaOH is defined elsewhere

    
        # Set color and line width
        color = "black" if int(np.round(x * 100)) % 10 == 0 else "grey"
        lw = 1.0 if color == "black" else 0.25
        
        # Plotting based on conditions
        temp_plot       = -1/temperaturesK     if invT else temperaturesK
        temp_plot_cryst = -1/(cryst_T+273.15)  if invT else cryst_T+273.15
        ylabel = 'Saturation Pressure [Pa]' if log else 'Saturation Pressure [Pa]'
        xlabel = 'Temperature [°C] (scaled -1/T)' if invT else 'Temperatur [°C]'
        
        # Plot isosteres
        if log:
            plt.semilogy(temp_plot, p, color=color, lw=lw)
        else:
            plt.plot(temp_plot, p, color=color, lw=lw)
        
        plt.ylabel(ylabel)
        plt.xlabel(xlabel)
        plt.xticks(-1/plotTemperatures if invT else plotTemperatures, plotTemperatures - 273.15)
        
        # Label concentrations
        if show_percentages and int(np.round(x * 100)) % 10 == 0 and x>0.29:
            label_pos = temp_plot[-1] + 1e-5 if invT else temp_plot[-1]+2
            plt.text(label_pos, p[-1], f'{x * 100:.0f} %', fontsize=8, color='black')

    # Plotting water line
    if log:
        plt.semilogy(temp_plot, waterPressure, color="grey", linestyle='--',
                     label='Pure Water')
    else:
        plt.plot(temp_plot, waterPressure, color="grey", linestyle='--', label='Pure Water')

    # Plotting crystallization curve
    if log:
        plt.semilogy(temp_plot_cryst, cryst_p, color="gray", linestyle='-',
                    lw=1.0, label='Crystallization Curve', zorder=101)
        # fill area between crystallization curve and the minimum positive value (e.g., 1)
        plt.fill_between(temp_plot_cryst, cryst_p, 1, where=(cryst_p > 1), color='white', zorder=100)
    else:
        plt.plot(temp_plot_cryst, cryst_p, color="gray", linestyle='-',
                 lw=1.0, label='Crystallization Curve', zorder=101)
        # fill area between crystallization curve and a fixed value (e.g., 50)
        plt.fill_between(temp_plot_cryst, cryst_p, 1, where=(cryst_p > 1),  color='white', zorder=100)

    

    # Setting axis limits
    if invT:
        plt.xlim(-1/temperaturesK[0], -1/temperaturesK[-1])
    else:
        plt.xlim(temperaturesK[0], temperaturesK[-1])
    
    if log:
        plt.ylim(220, 1.1e5)
    else:
        plt.ylim(0, max(waterPressure) * 1.1)  # Adjust as needed to make sure all data is visible


    plt.legend()
    
    if not editablePlot:
        plt.show()

    # turn on print statements again
    sys.stdout = stdout_old

# DONE
def solubility_temperature(xs):
    """
    Computes the solubility boundary temperature with salt concentration xs.

    Parameters:
        xs (array-like): Concentration of solution (0...1) kgSalt/kgSolution (column vector).

    Returns:
        ts (numpy array): Temperature of crystallization (column vector).

    Restrictions:
        - The concentration of salt must be between 0.0 and 0.78.
    """

    if np.any(xs < 0) or np.any(xs > 0.78):
        raise ValueError("The concentration of salt must be between 0.0 and 0.78.")

    # Coefficients
    A0 = np.array([0.422088, -0.378950, -0.519970, -1.149044, -2.385836, -2.807560])
    A1 = np.array([-0.066933, 3.456900, 3.400970, 5.509111, 8.084829, 4.678250])
    A2 = np.array([-0.282395, -3.531310, -2.851290, -4.642544, -5.303476, 0.000000])
    A3 = np.array([-355.514247])

    # Reduced temperature
    TcH2O = 647.26  # K

    # Initialize temperature array
    if isinstance(xs, (list, np.ndarray)):
        t = np.ones((6, len(xs)))
    else:
        t = np.ones((6, 1))

    # Iceline
    k = 0
    theta = A0[k] + A1[k] * xs + A2[k] * xs ** 2.0 + A3[k] * xs ** 7.5
    t[k] = theta * TcH2O - 273.15

    # Hydrates
    for k in range(1, 6):
        theta = A0[k] + A1[k] * xs + A2[k] * xs ** 2.0
        t[k] = theta * TcH2O - 273.15

    ts = np.max(t, axis=0)

    return ts

# DONE
def crystallization_curve(return_data=False):
    """    
    Author: Dorian Höffner
    Date: 2024-11-20
    Publication: Conde 2009 "Aqueous solutions of lithium and calcium chlorides: property formulations for use in air conditioning equipment design"

    Parameters:
        return_data (bool): If True, the data will be returned as a list of lists.

    Returns:
        None or list of lists: If return_data is True, the data will be returned as a list of lists (x, T).
    """

    # cryst_data = x, T // create cryst_data based on "solubility_temperate"
    xs_array = np.linspace(0.01, 0.78, 100)
    cryst_data = [[xs, solubility_temperature(xs)] for xs in xs_array]
    temperatures = [item[1] for item in cryst_data]
    
        
    if return_data:
        return cryst_data
    
    else:
        cryst_data = [[xs_array[i]*100, temperatures[i]] for i,_ in enumerate(xs_array)]
        # prepare plot
        plt.figure(figsize=(6,4), dpi=300)
        plt.plot([x[0] for x in cryst_data],
                    [x[1] for x in cryst_data],
                    label='Crystallization Curve',
                    color="black")
        plt.xlabel(r'$\mathrm{CaCl_2}$ Concentration [%]')
        plt.ylabel('Temperature [°C]')
        plt.xlim(0, 75)
        plt.ylim(-60, 200)
        # make beautiful grid
        plt.grid(True)
        plt.minorticks_on()
        plt.grid(which='major', linestyle='-', linewidth='0.2', color='black')
        plt.grid(which='minor', linestyle=':', linewidth='0.1', color='black')
        # add legend (top left)
        plt.legend(loc='upper left')


import numpy as np
import matplotlib.pyplot as plt
from scipy.optimize import fsolve

from pyXSteam.XSteam import XSteam
steamTable = XSteam(XSteam.UNIT_SYSTEM_MKS)  # m/kg/sec/°C/bar/W

import sys
import os

# NaOH property functions
def saturation_temperature(x, P):
    """
    Calculates the boiling point temperature of aqueous NaOH-H2O solutions at high concentrations.
    Uses saturation_pressure() to calculate the pressure and then calculates the temperature.
    ---
    Parameters:
        x (float): Mass fraction, defined as m_NaOH / (m_H2O + m_NaOH).
        P (float): Pressure in [Pa].
    ---
    Returns:
        float: Boiling point temperature in [°C].
    ---
    Author: Dorian Höffner
    Berechnung nach J. Olsson et. al "Thermophysical Properties of Aqueous NaOH-H2O Solutions at High Concentrations",
    International Journal of Thermophysics, Vol. 18, No. 3, 1997
    Boiling point temperature of aqueous NaOH-H2O Solution in °C
    """

    T_guess = 50 # initial guess for fsolve
    # suppress print statements
    with open(os.devnull, 'w') as devnull:
        old_stdout = sys.stdout
        sys.stdout = devnull
        T = fsolve(lambda T: saturation_pressure(x, T) - P, T_guess)[0]
        sys.stdout = old_stdout
    

    # implement checks for input values (checks are from saturation_pressure())
    if T < 0:
        print(f"Warning: values are outside the valid range, because T < 0 °C\nx = {x}\np = {P} Pa\nT = {T} °C")
    elif T < 20:
        if 1-x < 0.582:
            print(f"Warning: values are outside the valid range for T < 20 °C\nx = {x}\np = {P} Pa\nT = {T} °C")
    elif T < 60:
        if 1-x < 0.500:
            print(f"Warning: values are outside the valid range for 20 <= T < 60 °C\nx = {x}\np = {P} Pa\nT = {T} °C")
    elif T < 70:
        if 1-x < 0.353:
            print(f"Warning: values are outside the valid range for 60 <= T < 70 °C\nx = {x}\np = {P} Pa\nT = {T} °C")
    elif T < 150:
        if 1-x < 0.300:
            print(f"Warning: values are outside the valid range for 70 <= T < 150 °C\nx = {x}\np = {P} Pa\nT = {T} °C")
    elif T <= 200:
        if 1-x < 0.200:
            print(f"Warning: values are outside the valid range for 150 <= T <= 200 °C\nx = {x}\np = {P} Pa\nT = {T} °C")
                
    return T

def enthalpy(x, T):
    """
    Last Change: brought to python by Dorian Höffner
    Autor: 	Roman Ziegenhardt
    Quelle: 	Thermophysical Properties of Aqueous NaOH-H2O Solutions at High Concentrations, J. OIsson, A. Jernqvist, G. Aly, 
    		International Journal of Thermophysics Vol. 18. No. 3. 1997
    zuletzt geändert: Elisabeth Thiele: Temperatur in °C und salt mass
    fraction statt water mass fraction
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

    Parameter:
     T in [°C]
     x = m_NaOH/(m_H2O + m_NaOH)
     h in [kJ/kg]

    Wertebereich: t in Grad Celsius!
    0<=t<4			0.780 <=xi<=1
    4=<t<10	        0.680<=xi<=1
    10=<t<15		0.580<=xi<=1
    15=<t<26		0.540<=xi<=1
    26 =<t<37		0.440<=xi<=1
    37=<t<48		0.400<=xi<=1
    48=<t<60		0.340<=xi<=1
    60=<t<71		0.300<=xi<=1
    71=<t<82		0.280<=xi<=1
    82=<t<93		0.240<=xi<=1
    93=< t=<204	0.220<=xi<=1 """

    # Convert NaOH mass fraction to water mass fraction
    xi = 1 - x

    # implement checks for input values (checks are from saturation_pressure())
    if T < 0:
        raise ValueError('Temperature must be greater than 0 °C')
    elif T < 4:
        if xi < 0.780:
            print(f"Warning: values are outside the valid range for T < 4 °C\nx = {x}\nT = {T} °C")
    elif T < 10:
        if xi < 0.680:
            print(f"Warning: values are outside the valid range for 4 <= T < 10 °C\nx = {x}\nT = {T} °C")
    elif T < 15:
        if xi < 0.580:
            print(f"Warning: values are outside the valid range for 10 <= T < 15 °C\nx = {x}\nT = {T} °C")
    elif T < 26:
        if xi < 0.540:
            print(f"Warning: values are outside the valid range for 15 <= T < 26 °C\nx = {x}\nT = {T} °C")
    elif T < 37:
        if xi < 0.440:
            print(f"Warning: values are outside the valid range for 26 <= T < 37 °C\nx = {x}\nT = {T} °C")
    elif T < 48:
        if xi < 0.400:
            print(f"Warning: values are outside the valid range for 37 <= T < 48 °C\nx = {x}\nT = {T} °C")
    elif T < 60:
        if xi < 0.340:
            print(f"Warning: values are outside the valid range for 48 <= T < 60 °C\nx = {x}\nT = {T} °C")
    elif T < 71:
        if xi < 0.300:
            print(f"Warning: values are outside the valid range for 60 <= T < 71 °C\nx = {x}\nT = {T} °C")
    elif T < 82:
        if xi < 0.280:
            print(f"Warning: values are outside the valid range for 71 <= T < 82 °C\nx = {x}\nT = {T} °C")
    elif T < 93:
        if xi < 0.240:
            print(f"Warning: values are outside the valid range for 82 <= T < 93 °C\nx = {x}\nT = {T} °C")
    elif T <= 204:
        if xi < 0.220:
            print(f"Warning: values are outside the valid range for 93 <= T <= 204 °C\nx = {x}\nT = {T} °C")
    elif T>204:
        raise ValueError('Temperature must be less than 204 °C')

    # Coefficients
    k = np.array([1288.4485, -0.49649131, -4387.8908, -4.0915144, 4938.2298, 7.2887292, -1841.1890, -3.0202651])
    l = np.array([2.3087919, -9.0004252, 167.59914, -1051.6368, 3394.3378, -6115.0986, 6220.8249, -3348.8098, 743.87432])
    m = np.array([0.02302860, -0.37866056, 2.4529593, -8.2693542, 15.728833, -16.944427, 9.6254192, -2.2410628])
    n = np.array([-8.5131313e-5, 136.52823e-5, -875.68741e-5, 2920.0398e-5, -5488.2983e-5, 5841.8034e-5, -3278.7483e-5, 754.45993e-5])

    # Calculation of coefficients
    c1 = (k[0] + k[2]*xi + k[4]*(xi**2) + k[6]*(xi**3)) / (1 + k[1]*xi + k[3]*(xi**2) + k[5]*(xi**3) + k[7]*(xi**4))
    c2 = l[0] + l[1]*xi + l[2]*(xi**2) + l[3]*(xi**3) + l[4]*(xi**4) + l[5]*(xi**5) + l[6]*(xi**6) + l[7]*(xi**7) + l[8]*(xi**8)
    c3 = m[0] + m[1]*xi + m[2]*(xi**2) + m[3]*(xi**3) + m[4]*(xi**4) + m[5]*(xi**5) + m[6]*(xi**6) + m[7]*(xi**7)
    c4 = n[0] + n[1]*xi + n[2]*(xi**2) + n[3]*(xi**3) + n[4]*(xi**4) + n[5]*(xi**5) + n[6]*(xi**6) + n[7]*(xi**7)

    # Calculate enthalpy
    h = c1 + c2*T + c3*(T**2) + c4*(T**3)

    return float(h)

def saturation_pressure(xi, T):
    """
    Calculates the pressure of aqueous NaOH-H2O solutions at high concentrations.
    
    Last Change: Dorian Höffner 2024-04-26 (translated to Python, changed input T to [°C], changed return value to [Pa])
    Author: Anna Jahnke, Roman Ziegenhardt
    Source: Thermophysical Properties of Aqueous NaOH-H2O Solutions at High Concentrations, J. Olsson, A. Jernqvist, G. Aly,
            International Journal of Thermophysics Vol. 18, No. 3, 1997
    
    Parameters:
    T (array-like): Temperature in [°C].
    x (array-like): Mass fraction, defined as m_NaOH / (m_H2O + m_NaOH).
    
    Returns:
    p (float or array-like): Pressure in [Pa]

    Notes:
    Wertebereich: t in Grad Celsius!
    0<=t<20    		0.582<=1-x<=1
    20<=t<60		0.500<=1-x<=1
    60<=t<70       	0.353<=1-x<=1
    70<=t<150		0.300<=1-x<=1
    150<=t<=200		0.200<=1-x<=1
    """

    # convert NaOH mass fraction to water mass fraction
    x = 1 - xi

    # implement checks for input values (checks are from saturation_pressure())
    if T < 0:
        print(f"Warning: values are outside the valid range for T < 0 °C\nx = {xi}\nT = {T} °C")
    elif T < 20:
        if x <= 0.582:
            print(f"Warning: values are outside the valid range for T < 20 °C\nx = {xi}\nT = {T} °C")
    elif T < 60:
        if x <= 0.500:
            print(f"Warning: values are outside the valid range for 20 <= T < 60 °C\nx = {xi}\nT = {T} °C")
    elif T < 70:
        if x <= 0.353:
            print(f"Warning: values are outside the valid range for 60 <= T < 70 °C\nx = {xi}\nT = {T} °C")
    elif T < 150:
        if x <= 0.300:
            print(f"Warning: values are outside the valid range for 70 <= T < 150 °C\nx = {xi}\nT = {T} °C")
    elif T <= 200:
        if x <= 0.200:
            print(f"Warning: values are outside the valid range for 150 <= T <= 200 °C\nx = {xi}\nT = {T} °C")

   
    if isinstance(x, (int, float)) or isinstance(T, (int, float)):
        pass
    elif len(x) != len(T):
        raise ValueError('x and T must have the same length')

    # Constants
    k = np.array([-113.93947, 209.82305, 494.77153, 6860.8330, 2676.6433,
                  -21740.328, -34750.872, -20122.157, -4102.9890])
    l = np.array([16.240074, -11.864008, -223.47305, -1650.3997, -5997.3118,
                  -12318.744, -15303.153, -11707.480, -5364.9554, -1338.5412,
                  -137.96889])
    m = np.array([-226.80157, 293.17155, 5081.8791, 36752.126, 131262.00,
                  259399.54, 301696.22, 208617.90, 81774.024, 15648.526,
                  906.29769])

    # Calculate pressure
    log_x = np.log(x)
    a1 = np.polyval(k[::-1], log_x)
    a2 = np.polyval(l[::-1], log_x)
    a3 = np.polyval(m[::-1], log_x)

    logP = (a1 + a2 * T) / (T - a3)
    p = np.exp(logP) * 1000

    return float(p) # pressure in Pa

def saturation_concentration(p, T):
    """
    Calculates the saturation concentration of NaOH in water based on the temperature and pressure.

    Author: Dorian Höffner
    Date: 2024-09-17

    Parameters:
        p (float): Pressure in Pa.
        T (float): Temperature in °C.

    Returns:
        float: Saturation concentration in kg NaOH / kg solution.
    """

    # Calculate the saturation concentration
    with open(os.devnull, 'w') as devnull:
        old_stdout = sys.stdout
        sys.stdout = devnull
        x = 0.0
        while saturation_pressure(x, T) - p > 1:   # 1 Pa tolerance
            x += 0.001
        sys.stdout = old_stdout


    # implement checks for input values (checks are from saturation_pressure())
    if T < 0:
        raise ValueError('Temperature must be greater than 0 °C')
    elif T < 20:
        if 1-x < 0.582:
            print(f"Warning: values are outside the valid range for T < 20 °C\nx = {x}\np = {P} Pa\nT = {T} °C")
    elif T < 60:
        if 1-x < 0.500:
            print(f"Warning: values are outside the valid range for 20 <= T < 60 °C\nx = {x}\np = {P} Pa\nT = {T} °C")
    elif T < 70:
        if 1-x < 0.353:
            print(f"Warning: values are outside the valid range for 60 <= T < 70 °C\nx = {x}\np = {P} Pa\nT = {T} °C")
    elif T < 150:
        if 1-x < 0.300:
            print(f"Warning: values are outside the valid range for 70 <= T < 150 °C\nx = {x}\np = {P} Pa\nT = {T} °C")
    elif T <= 200:
        if 1-x < 0.200:
            print(f"Warning: values are outside the valid range for 150 <= T <= 200 °C\nx = {x}\np = {P} Pa\nT = {T} °C")

    return x

def density(x, T):
    """
    DensityNaOH calculates the density of NaOH solution based on concentration and temperature.
    ---
    Parameters:
        x (array-like): Mole fraction of NaOH in the solution. [m_NaOH / (m_h2o+m_naoh)]
        T (array-like): Temperature in °C.
    ---
    Returns:
        rho (numpy array): Density in kg/m^3.
    ---
    Source: "Thermophysical Properties of Aqueous NaOH-H2O Solutions at High Concentrations,
    J. OIsson, A. Jernqvist, G. Aly, International Journal of Thermophysics Vol. 18. No. 3. 1997"
    ---
    Restrictions: 
     0 <= t <  10 ------ 0.0 < x <= 0.2
    10 <= t <  20 ------ 0.0 < x <= 0.3
    20 <= t <  60 ------ 0.0 < x <= 0.5
    60 <= t <  70 ------ 0.0 < x <= 0.6
    70 <= t < 150 ------ 0.0 < x <= 0.7
    150 <= t < 200 ------ 0.0 < x <= 0.8
    """

    # check validity
    if T < 0:
        raise ValueError('Temperature must be greater than 0 °C')
    elif T < 10:
        if x > 0.2:
            print(f"Warning: values are outside the valid range for T < 10 °C\nx = {x}\nT = {T} °C")
    elif T < 20:
        if x > 0.3:
            print(f"Warning: values are outside the valid range for 10 <= T < 20 °C\nx = {x}\nT = {T} °C")
    elif T < 60:
        if x > 0.5:
            print(f"Warning: values are outside the valid range for 20 <= T < 60 °C\nx = {x}\nT = {T} °C")
    elif T < 70:
        if x > 0.6:
            print(f"Warning: values are outside the valid range for 60 <= T < 70 °C\nx = {x}\nT = {T} °C")
    elif T < 150:
        if x > 0.7:
            print(f"Warning: values are outside the valid range for 70 <= T < 150 °C\nx = {x}\nT = {T} °C")
    elif T <= 200:
        if x > 0.8:
            print(f"Warning: values are outside the valid range for 150 <= T <= 200 °C\nx = {x}\nT = {T} °C")
    elif T > 200:
        raise ValueError('Temperature must be less than 200 °C')
    

    # transform x=x_NaOH to x=x_h2o
    if isinstance(x, (int, float)) or isinstance(T, (int, float)):
        x = 1 - x
        pass
    elif len(x) != len(T):
        raise ValueError('x and T must have the same length')
    else:
        x = 1 - np.array(x)
        T = np.array(T)

    # Coefficients
    k = np.array([5007.2279636, -25131.164248, 74107.692582, -104657.48684, 69821.773186, -18145.911810])
    l = np.array([-64.786269079, 525.34360564, -1608.4471903, 2350.9753235, -1660.9035108, 457.6437435])
    m = np.array([0.24436776978, -1.9737722344, 6.04601497138, -8.9090614947, 6.37146769397, -1.7816083111])

    x = np.array(x)
    
    b1 = k[0] + k[1] * (x ** (1/2)) + k[2] * x + k[3] * (x ** (3/2)) + k[4] * (x ** 2) + k[5] * (x ** (5/2))
    b2 = l[0] + l[1] * (x ** (1/2)) + l[2] * x + l[3] * (x ** (3/2)) + l[4] * (x ** 2) + l[5] * (x ** (5/2))
    b3 = m[0] + m[1] * (x ** (1/2)) + m[2] * x + m[3] * (x ** (3/2)) + m[4] * (x ** 2) + m[5] * (x ** (5/2))

    rho = b1 + b2 * T + b3 * (T ** 2)

    return float(rho)

def specific_heat_capacity(x, T):
    """
    Calculate the specific heat capacity of a NaOH solution as a function of temperature and concentration.
    The function is based on the following publication:
    Alexandrov 2004 - "The Equations for Thermophysical Properties of Aqueous Solutions of Sodium Hydroxide"
    ---
    x: float
        Concentration of NaOH in the solution [kg/kg]
    T: float
        Temperature of the solution [°C]
    ---
    returns: float
        Specific heat capacity of the solution [kJ/kgK]
    ---
    Author: Dorian Höffner
    Date: 2024-08-19
    ---
    Restrictions:
    0 <= T <= 275.85 °C
    0 <= x <= 0.16 kg/kg
    """

    # check validity
    if T < 0:
        raise ValueError('Temperature must be greater than 0 °C')
    elif T > 275.85:
        raise ValueError('Temperature must be less than 275.85 °C')
    elif x < 0:
        raise ValueError('Concentration must be greater than 0 kg/kg')
    elif x > 0.16:
        print(f"Warning: Values are outside the valid range for x > 0.16 kg/kg")

    # Constants
    a01=9.8555259e1
    a11=-3.4501318e2
    a21=4.8180532e2
    a31=-3.3440616e2
    a41=1.1516735e2
    a51=-1.5708814e1

    a02=-3.4357815e1
    a12=1.1674552e2
    a22=-1.5776854e2
    a32=1.0577045e2
    a42=-3.5099188e1
    a52=4.5935013
    
    a03=1.9791083
    a13=-5.3828966
    a23=5.5124212
    a33=-2.5430046
    a43=4.5943595e-1
    
    a04=-5.6191575e-2
    a14=8.5936388e-2
    a24=-1.6966718e-2
    a34=-1.4864492e-2
    
    a05=7.9944152e-3
    a15=-1.5444457e-2
    a25=7.5030322e-3


    cp_water = steamTable.CpL_t(T) # [kJ/kgK]

    # calculate molality (mol/kg) from concentration (kg/kg)
    M = 39.9971e-3          # [kg/mol]
    m = x / M               # [mol/kg]

    # calculate relative temperature
    t = (T + 273.15) / 273.15

    cp_diff = m    * (a01 + a11 * t + a21 * t**2 + a31 * t**3 + a41 * t**4 + a51 * t**5) + \
              m**2 * (a02 + a12 * t + a22 * t**2 + a32 * t**3 + a42 * t**4 + a52 * t**5) + \
              m**3 * (a03 + a13 * t + a23 * t**2 + a33 * t**3 + a43 * t**4) + \
              m**4 * (a04 + a14 * t + a24 * t**2 + a34 * t**3) + \
              m**5 * (a05 + a15 * t + a25 * t**2)

    if x <= 0.0001:
        cp = cp_water
    else:
        cp = cp_water - cp_diff * 1e-3 # [kJ/kgK]

    return cp

def dynamic_viscosity(x, T, p):
    """
    Calculate the dynamic viscosity of a NaOH solution as a function of temperature and concentration.
    The function is based on the following publication:
    Alexandrov 2004 - "The Equations for Thermophysical Properties of Aqueous Solutions of Sodium Hydroxide"
    ---
    x: float
        Concentration of NaOH in the solution [kg/kg]
    T: float
        Temperature of the solution [°C]
    p: float
        Pressure of the solution [Pa]
    ---
    returns: float
        Dynamic viscosity of the solution [Pa s]
    ---
    Restrictions:
    0 <= T <= 250.85 °C
    0 <= x <= 0.12 kg/kg
    0 <= p <= 7 MPa
    """

    # check validity
    if T < 0:
        raise ValueError('Temperature must be greater than 0 °C')
    elif T > 250.85:
        raise ValueError('Temperature must be less than 250.85 °C')
    elif x < 0:
        raise ValueError('Concentration must be greater than 0 kg/kg')
    elif p < 0:
        raise ValueError('Pressure must be greater than 0 Pa')
    elif p > 7e6:
        print(f"Warning: Values are outside the valid range for p > 15 MPa")
    elif x > 0.12:
        print(f"Warning: Values are outside the valid range for x > 0.12 kg/kg")


    # prepare inputs
    M = 39.9971e-3                      # [kg/mol]
    m = x / M                           # [mol/kg]
    T0 = 293.15                         # [K] (yes 293.15, not 273.15)
    t = T0 / (T + 273.15)               # [-]
    t1 = t - 1                          # [-]

    p = p / 1e6                         # [Pa]  -> [MPa]
    p_sw = steamTable.psat_t(T) / 1e1   # [bar] -> [MPa]
    
    # coeffs from cited publicatiosn
    c = np.array([np.nan, 5.17341030, 9.81838817, 2.83021985e1, 7.02071954e1, -9.92041252e2, -1.13267055e4, -5.10988292e4, -1.18863488e5, -1.41053273e5, -6.78490604e4])
    d = np.array([-3.18833435e-1, -1.07314454e1, -8.61347656e1, -6.50268842e2, -6.06767730e3, -4.07022741e4, -1.59650983e5, -3.53438962e5, -4.11357235e5, -1.96118714e5, np.nan])

    expression1 =        c[1] * t1 + c[2] * t1**2 + c[3] * t1**3 + c[4] * t1**4 + c[5] * t1**5 + c[6] * t1**6 + c[7] * t1**7 + c[8] * t1**8 + c[9] * t1**9 + c[10] * t1**10
    expression2 = d[0] + d[1] * t1 + d[2] * t1**2 + d[3] * t1**3 + d[4] * t1**4 + d[5] * t1**5 + d[6] * t1**6 + d[7] * t1**7 + d[8] * t1**8 + d[9] * t1**9
    my_water    = 1001.6 * (t1 + 1)**2 * np.exp(expression1) + (p-p_sw) * expression2 # [µPa s]

    # coeffs from actual publication
    b11 =  5.7070102e-1
    b21 =  4.9395013e-1
    b31 = -2.0417183
    b41 =  1.1654862

    b12 = -2.9922166e-1
    b22 =  3.7957782e-1
    b32 = -7.423751e-2

    b13 =  4.9815412e-2
    b23 = -4.8332728e-2

    expression3 = m    * (b11 * t + b21 * t**2 + b31 * t**3 + b41 * t**4) + \
                  m**2 * (b12 * t + b22 * t**2 + b32 * t**3) + \
                  m**3 * (b13 * t + b23 * t**2)
    
    my_rel = np.exp(expression3)

    my = my_water * np.exp(expression3) # [µPa s]

    return float(my * 1e-6) # [Pa s]

def thermal_conductivity(x, T, p):
    """
    Calculate the thermal conductivity of a NaOH solution as a function of temperature and concentration.
    The function is based on the following publication:
    Alexandrov 2004 - "The Equations for Thermophysical Properties of Aqueous Solutions of Sodium Hydroxide"
    ---
    x: float
        Concentration of NaOH in the solution [kg/kg]
    T: float
        Temperature of the solution [°C]
    ---
    returns: float
        Thermal conductivity of the solution [W/mK]
    ---
    Restrictions:
        0 <= T <= 132.85 °C
        0 <= x <= 0.2 kg/kg
        0 <= p <= 15 MPa
    """

    # check validity
    if T < 0:
        raise ValueError('Temperature must be greater than 0 °C')
    elif T > 132.85:
        raise ValueError('Temperature must be less than 132.85 °C')
    elif x < 0:
        raise ValueError('Concentration must be greater than 0 kg/kg')
    elif p < 0:
        raise ValueError('Pressure must be greater than 0 Pa')
    elif p > 15e6:
        raise ValueError('Pressure must be less than 15 MPa')
    elif x > 0.2:
        print(f"Warning: Values are outside the valid range for x > 0.2 kg/kg")

    # prepare inputs for water calculation
    p_sw = steamTable.psat_t(T) / 1e1   # [bar] -> [MPa]
    p = p / 1e6                         # [Pa]  -> [MPa]
    T02 = 273.15 + 20                   # [K]
    t = T02 / (T + 273.15) - 1          # [-]

    # coeffs from cited publication (for water)
    g = np.array([5.99454842e-1, -4.82554378e-1, -4.31229616e-1, -8.62555022e-1, -3.80050418e-1, 4.85828450e1, 3.35400696e2, 1.08007806e3, 1.67727081e3, 1.04225629e3])
    q = np.array([5.31492446e-4, 3.46658996e-4, 1.23050434e-2, 1.27873471e-1, -7.40820487e-1, -1.93072528e1, -1.22835056e2, -3.66150909e2, -5.31321978e2, -3.03153185e2])

    expression1 = g[0] + g[1] * t**1 + g[2] * t**2 + g[3] * t**3 + g[4] * t**4 + g[5] * t**5 + g[6] * t**6 + g[7] * t**7 + g[8] * t**8 + g[9] * t**9
    expression2 = q[0] + q[1] * t**1 + q[2] * t**2 + q[3] * t**3 + q[4] * t**4 + q[5] * t**5 + q[6] * t**6 + q[7] * t**7 + q[8] * t**8 + q[9] * t**9
    lambda_water = expression1 + (p-p_sw) * expression2  # [kW/mK]

    
    # prepare inputs for NaOH calculation
    T0 = 403.0 # [K]
    t1 = (T+273.15) / T0
    M = 39.9971e-3          # [kg/mol]
    m = x / M               # [mol/kg]
    
    # coeffs from actual publication (for NaOH)
    e01 = 3.2900544e-1
    e11 = -1.1048583
    e21 = 1.2503803
    e31 = -4.4228179e-1

    e02 = -2.1990820e-2
    e12 = 5.9100989e-2
    e22 = -4.4407173e-2
    
    e03 = 1.5069324e-3
    e13 = -4.3273501e-3
    e23 = 3.3763248e-3

    expression3 = m    * (e01 + e11 * t1 + e21 * t1**2 + e31 * t1**3) + \
                  m**2 * (e02 + e12 * t1 + e22 * t1**2) + \
                  m**3 * (e03 + e13 * t1 + e23 * t1**2)
    
    lambda_NaOH = lambda_water + expression3 # [kW/mK]

    return float(lambda_NaOH * 1e3) # [W/mK]

def dhdx(x,T):
    # Author: Elisabeth Thiele
    # brought to python by Dorian Höffner
    # partielles Differential dh/dx

    # Variables:
    # T:        Temperatur            in °C
    # x:        NaOH concentration in kg NaOH/kg solution
    # dhdx:     partielle Ableitung Enthalpie nach Massenanteil in [kJ/kg]

    delta = 0.000001
    x1 = x - delta
    x2 = x + delta

    dhdx = (enthalpy(x2,T) - enthalpy(x1,T)) / (2*delta)
    return dhdx

def dhdT(x,T):
    # Author: Elisabeth Thiele
    # partielles Differential dh/dx

    # Variables:
    # T:        Temperatur [°C]
    # x:        NaOH concentration [kg NaOH/kg solution]
    # dhdx:     partielle Ableitung Enthalpie nach Temperatur in [kJ/kgK] EINHEIT?

    delta = 0.0001
    T1 = T - delta
    T2 = T + delta

    dhdT = (enthalpy(x,T1) - enthalpy(x,T2)) / (2*delta)
    return dhdT


def pTDiagram(log=True, invT=True, editablePlot=False, show_percentages=True):
    """
    Plots the pressure-temperature diagram for NaOH-H2O solutions.
    Author: Dorian Höffner 2024-04-26
    Last Change: 2024-09-17
    
    Parameters:
        log (bool): If True, the y-axis will be logarithmic.
        invT (bool): If True, the x-axis will be scaled as -1/T.
        editablePlot (bool): If True, the plot will be editable.
        show_percentages (bool): If True, the concentrations will be labeled.

    Returns:
        None
    """

    # suppress all print statements
    devnull = open(os.devnull, 'w')
    stout_old = sys.stdout
    sys.stdout = devnull

    
    steamTable = XSteam(XSteam.UNIT_SYSTEM_MKS)  # m/kg/sec/°C/bar/W

    # get data to plot crystallization curve
    cryst_data = crystallization_curve(return_data=True)
    cryst_T = np.array([T for x, T in cryst_data])
    #cryst_p = np.array([saturation_pressure(x, T) for x, T in cryst_data])
    cryst_p = []
    for x, T in cryst_data:
        try:
            p = saturation_pressure(x, T)
        except:
            p = 0
        cryst_p.append(p)
    cryst_p = np.array(cryst_p)
    
    # Temperature range
    temperaturesC = np.arange(0, 100, 1)
    temperaturesK = temperaturesC + 273.15
    concentrations = np.arange(0.1, 0.751, 0.01)
    cr_temperaturesC = cryst_T
    cr_temperaturesK = cryst_T + 273.15
    
    # Prepare the plot
    plt.figure(dpi=300)
    
    # these temperatures are used for the x-axis
    plotTemperatures = np.arange(0, 101, 10) + 273.15
    
    # Calculate water vapor pressure using XSteam
    waterPressure = [steamTable.psat_t(T - 273.15) * 1e5 for T in temperaturesK]  # convert bar to Pa
    
    # SUPPRESS WARNINGS


    # Plot the data
    for x in concentrations:

        # Calculate saturation pressure for each temperature at the given concentration
        p = []
        for T in temperaturesC:
            p.append(saturation_pressure(x, T))           

    
        # Set color and line width
        color = "black" if int(np.round(x * 100)) % 10 == 0 else "grey"
        lw = 1.0 if color == "black" else 0.25
        
        # Plotting based on conditions
        temp_plot       = -1/temperaturesK     if invT else temperaturesK
        temp_plot_cryst = -1/(cryst_T+273.15)  if invT else cryst_T+273.15
        ylabel = 'Saturation Pressure [Pa]' if log else 'Saturation Pressure [Pa]'
        xlabel = 'Temperature [°C] (scaled -1/T)' if invT else 'Temperatur [°C]'
        
        # Plot isosteres
        if log:
            plt.semilogy(temp_plot, p, color=color, lw=lw)
        else:
            plt.plot(temp_plot, p, color=color, lw=lw)
        
        plt.ylabel(ylabel)
        plt.xlabel(xlabel)
        plt.xticks(-1/plotTemperatures if invT else plotTemperatures, plotTemperatures - 273.15)
        
        # Label concentrations
        if show_percentages and int(np.round(x * 100)) % 10 == 0:
            label_pos = temp_plot[-1] + 1e-5 if invT else temp_plot[-1]+2
            plt.text(label_pos, p[-1], f'{x * 100:.0f} %', fontsize=8, color='black')

    # Plotting water line
    if log:
        plt.semilogy(temp_plot, waterPressure, color="grey", linestyle='--',
                    label='Pure Water')
    else:
        plt.plot(temp_plot, waterPressure, color="grey", linestyle='--', label='Pure Water')

    # Plotting crystallization curve
    if log:
        plt.semilogy(temp_plot_cryst, cryst_p, color="gray", linestyle='-',
                    lw=1.0, label='Crystallization Curve', zorder=101)
        # fill area between crystallization curve and the minimum positive value (e.g., 1)
        plt.fill_between(temp_plot_cryst, cryst_p, 1, where=(cryst_p > 1), color='white', zorder=100)
    else:
        plt.plot(temp_plot_cryst, cryst_p, color="gray", linestyle='-',
                lw=1.0, label='Crystallization Curve', zorder=101)
        # fill area between crystallization curve and a fixed value (e.g., 50)
        plt.fill_between(temp_plot_cryst, cryst_p, 1, where=(cryst_p > 1),  color='white', zorder=100)

    

    # Setting axis limits
    if invT:
        plt.xlim(-1/temperaturesK[0], -1/temperaturesK[-1])
    else:
        plt.xlim(temperaturesK[0], temperaturesK[-1])
    
    if log:
        plt.ylim(50, 1e5)
    else:
        plt.ylim(0, max(waterPressure) * 1.1)  # Adjust as needed to make sure all data is visible


    plt.legend()
    
    if not editablePlot:
        plt.show()

    # reset print statements
    sys.stdout = stout_old
        

def solubility_temperature(x):
    """
    Return the crystallization temperature of NaOH solution in water.
    Source: Wang et al. 2008 (data was extracted from the plot)
    ---
    Parameters:
        x (float or np.array): Concentration of NaOH in kg/kg
    ---
    Returns:
        float or np.array: Crystallization temperature in °C
    ---
    Restrictions:
        0 <= x <= 0.787 kg/kg
    """

    # check validity
    if x < 0:
        raise ValueError('Concentration must be greater than 0 kg/kg')
    elif x > 0.787:
        raise ValueError('Concentration x = {x} must be less than 0.787 kg/kg')

    # cryst_data = x, T
    cryst_data   = [[0.2707275803722504,0.0],
                    [2.165820642978004,-1.3333333333333357],
                    [4.060913705583756,-2.933333333333337],
                    [5.685279187817259,-4.533333333333331],
                    [7.580372250423012,-6.666666666666664],
                    [9.34010152284264,-8.533333333333335],
                    [10.96446700507614,-10.933333333333334],
                    [12.588832487309643,-13.333333333333332],
                    [14.077834179357025,-16.0],
                    [15.5668358714044,-18.933333333333334],
                    [16.785109983079526,-21.866666666666667],
                    [17.868020304568528,-24.8],
                    [18.68020304568528,-28.0],
                    [20.169204737732656,-26.4],
                    [21.658206429780037,-24.8],
                    [22.74111675126904,-22.4],
                    [23.688663282571916,-20.53333333333333],
                    [24.771573604060915,-18.666666666666668],
                    [25.583756345177665,-14.133333333333333],
                    [26.395939086294412,-9.6],
                    [27.34348561759729,-5.333333333333336],
                    [28.56175972927242,-1.3333333333333357],
                    [29.780033840947542,1.8666666666666671],
                    [31.810490693739425,5.066666666666663],
                    [32.89340101522843,7.733333333333334],
                    [34.247038917089675,10.133333333333333],
                    [35.736040609137056,12.533333333333331],
                    [37.36040609137056,14.666666666666664],
                    [38.984771573604064,15.733333333333334],
                    [40.33840947546531,15.466666666666669],
                    [41.96277495769881,13.866666666666667],
                    [43.45177664974619,11.733333333333334],
                    [44.67005076142132,8.799999999999997],
                    [45.6175972927242,6.133333333333333],
                    [47.10659898477157,8.0],
                    [48.73096446700507,9.866666666666667],
                    [50.35532994923858,11.733333333333334],
                    [51.43824027072758,12.533333333333331],
                    [52.25042301184433,17.333333333333336],
                    [53.06260575296108,22.133333333333333],
                    [53.87478849407783,26.66666666666667],
                    [54.95769881556684,32.0],
                    [55.90524534686971,36.8],
                    [56.98815566835872,41.06666666666666],
                    [58.34179357021996,45.86666666666666],
                    [59.96615905245347,50.400000000000006],
                    [61.86125211505922,54.93333333333334],
                    [63.756345177664976,58.66666666666667],
                    [65.65143824027072,61.33333333333333],
                    [68.08798646362098,63.46666666666667],
                    [70.25380710659898,64.0],
                    [72.14890016920474,63.73333333333333],
                    [73.36717428087987,62.93333333333334],
                    [73.63790186125212,68.53333333333333],
                    [74.04399323181049,76.26666666666667],
                    [74.45008460236886,83.2],
                    [74.72081218274111,89.06666666666666],
                    [75.26226734348562,96.53333333333332],
                    [75.80372250423012,103.73333333333332],
                    [76.34517766497461,110.93333333333334],
                    [77.02199661590524,118.4],
                    [77.834179357022,126.13333333333333],
                    [78.78172588832487,133.33333333333334]]

    # create interpolation function
    x = x * 100  # convert to percentage
    t_solubility = np.interp(x, [x for x, T in cryst_data], [T for x, T in cryst_data])

    return t_solubility # [°C]


def crystallization_curve(return_data=False):
    """    
    Author: Dorian Höffner
    Date: 2024-09-17
    Data Source: Data was extracted from plot in publication.
    Publication: "Wang 2008 - Cellulose Fiber Dissolution in Sodium Hydroxide Solution at Low Temperature: Dissolution Kinetics and Solubility Improvement"

    Parameters:
        return_data (bool): If True, the data will be returned as a list of lists.

    Returns:
        None or list of lists: If return_data is True, the data will be returned as a list of lists (x, T).
    """
    
    # cryst_data = x, T
    cryst_data   = [[0.2707275803722504,0.0],
                    [2.165820642978004,-1.3333333333333357],
                    [4.060913705583756,-2.933333333333337],
                    [5.685279187817259,-4.533333333333331],
                    [7.580372250423012,-6.666666666666664],
                    [9.34010152284264,-8.533333333333335],
                    [10.96446700507614,-10.933333333333334],
                    [12.588832487309643,-13.333333333333332],
                    [14.077834179357025,-16.0],
                    [15.5668358714044,-18.933333333333334],
                    [16.785109983079526,-21.866666666666667],
                    [17.868020304568528,-24.8],
                    [18.68020304568528,-28.0],
                    [20.169204737732656,-26.4],
                    [21.658206429780037,-24.8],
                    [22.74111675126904,-22.4],
                    [23.688663282571916,-20.53333333333333],
                    [24.771573604060915,-18.666666666666668],
                    [25.583756345177665,-14.133333333333333],
                    [26.395939086294412,-9.6],
                    [27.34348561759729,-5.333333333333336],
                    [28.56175972927242,-1.3333333333333357],
                    [29.780033840947542,1.8666666666666671],
                    [31.810490693739425,5.066666666666663],
                    [32.89340101522843,7.733333333333334],
                    [34.247038917089675,10.133333333333333],
                    [35.736040609137056,12.533333333333331],
                    [37.36040609137056,14.666666666666664],
                    [38.984771573604064,15.733333333333334],
                    [40.33840947546531,15.466666666666669],
                    [41.96277495769881,13.866666666666667],
                    [43.45177664974619,11.733333333333334],
                    [44.67005076142132,8.799999999999997],
                    [45.6175972927242,6.133333333333333],
                    [47.10659898477157,8.0],
                    [48.73096446700507,9.866666666666667],
                    [50.35532994923858,11.733333333333334],
                    [51.43824027072758,12.533333333333331],
                    [52.25042301184433,17.333333333333336],
                    [53.06260575296108,22.133333333333333],
                    [53.87478849407783,26.66666666666667],
                    [54.95769881556684,32.0],
                    [55.90524534686971,36.8],
                    [56.98815566835872,41.06666666666666],
                    [58.34179357021996,45.86666666666666],
                    [59.96615905245347,50.400000000000006],
                    [61.86125211505922,54.93333333333334],
                    [63.756345177664976,58.66666666666667],
                    [65.65143824027072,61.33333333333333],
                    [68.08798646362098,63.46666666666667],
                    [70.25380710659898,64.0],
                    [72.14890016920474,63.73333333333333],
                    [73.36717428087987,62.93333333333334],
                    [73.63790186125212,68.53333333333333],
                    [74.04399323181049,76.26666666666667],
                    [74.45008460236886,83.2],
                    [74.72081218274111,89.06666666666666],
                    [75.26226734348562,96.53333333333332],
                    [75.80372250423012,103.73333333333332],
                    [76.34517766497461,110.93333333333334],
                    [77.02199661590524,118.4],
                    [77.834179357022,126.13333333333333],
                    [78.78172588832487,133.33333333333334]]
        
    if return_data:
        cryst_data = [[x/100, T] for x, T in cryst_data]
        return cryst_data
    
    else:
        # prepare plot
        plt.figure(figsize=(6,4), dpi=300)
        plt.plot([x[0] for x in cryst_data],
                    [x[1] for x in cryst_data],
                    label='Crystallization Curve',
                    color="black")
        plt.xlabel('NaOH Concentration [%]')
        plt.ylabel('Temperature [°C]')
        plt.xlim(0, 78.5)
        plt.ylim(-35, 140)
        # make beautiful grid
        plt.grid(True)
        plt.minorticks_on()
        plt.grid(which='major', linestyle='-', linewidth='0.2', color='black')
        plt.grid(which='minor', linestyle=':', linewidth='0.1', color='black')
        # add legend (top left)
        plt.legend(loc='upper left')


import numpy as np
from scipy.interpolate import griddata
from scipy.optimize import fsolve
import matplotlib.pyplot as plt

from pyXSteam.XSteam import XSteam
steamTable = XSteam(XSteam.UNIT_SYSTEM_MKS)  # m/kg/sec/°C/bar/W

import sys
import os
import contextlib

# LiCl property functions

# DONE
import sys
import contextlib
from scipy.optimize import fsolve

def saturation_temperature(x, p):
    """
    Calculates the saturation temperature of a LiCl solution based on the pressure and concentration.
    ---
    Parameters:
        x (float): Concentration of LiCl in the solution [kg/kg].
        p (float): Pressure in [Pa].
    ---
    author: Dorian Höffner 11/2024
    """

    # Silence print statements during fsolve
    T_guess = 20
    with contextlib.redirect_stdout(sys.stdout):  # Redirect stdout temporarily
        T = fsolve(lambda T: saturation_pressure(x, T) - p, T_guess)[0]

    # Check if the solution is in liquid state
    t_sol = solubility_temperature(x)
    if T < t_sol:
        print(f"Warning: The temperature {T} is below the solubility temperature {t_sol} at x={x}.")
    
    return T


# TODO
def enthalpy(x, T):
    """
    Last Change:
    Author:
    Quelle:

    Parameter:
     T in [°C]
     x = m_LiCl/(m_H2O + m_LiCl)
     h in [kJ/kg]
     """
    
    # sensible: use water as reference

    # excess enthalpy based on dilution ???
    print("Enthalpy function not available. 'differential_enthalpy_AD' is to be used instead.")

    #return h # enthalpy in [kJ/kg]

# DONE
def differential_enthalpy_AD(x, T):
    """
    Calculates the differential enthalpy [kJ/kgH2O] of a LiCl solution based on concentration and temperature.
    How much additional energy (in comparison to pure vaporization enthalpy) is needed to vaporize a certain amount of water in a LiCl solution.
    
    Parameters:
        x (float): Concentration of LiCl in the solution [kg/kg].
        T (float): Temperature of the solution [°C].

    Returns:
        float: Differential enthalpy [kJ/kgH2O].

    Author: Dorian Höffner 11/2024
    Source: CONDE2009 "Aqueous solutions of lithium and calcium chlorides: property formulations for use in air conditioning equipment design" (2009)
    """

    H1 = 0.845
    H2 = -1.965
    H3 = -2.265
    H4 = 0.6
    H5 = 169.105
    H6 = 457.850

    # Reduced Temperature (with critical T of water)
    Theta = (T + 273.15) / 647.1 # FLAG

    # zeta
    zeta = x / (H4-x)

    # Reference differental enthalpy
    dh_dil0 =  H5 + H6 * Theta

    # differental enthalpy
    dh_dil = dh_dil0 * (1 + (zeta/H1)**H2)**H3

    # check if the solution is in liquid state
    t_sol = solubility_temperature(x)
    if T < t_sol:
        print(f"Warning: The temperature {T} is below the solubility temperature {t_sol} at x={x}.")

    return dh_dil


# DONE
def saturation_pressure(x, T):
    """
    Computes the vapor pressure over a solution of LiCl with temperature T and concentration x.

    Parameters:
        T (numpy array): Temperature of the solution in [°C]
        x (numpy array): Concentration of the solution as m_LiCl / (m_H2O + m_LiCl)
    ---
    Returns:
        pd (float or array-like): Vapor pressure over the solution in [Pa]
    ---
    Last Change: Dorian Höffner 11/2024
    Author: O. Buchin 03/2011
    Source: CONDE2009 "Aqueous solutions of lithium and calcium chlorides: property formulations for use in air conditioning equipment design" (2009)
    """
    # Mass fraction of salt in solution
    zeta = x

    # Coefficients
    pi0 = 0.28
    pi1 = 4.3
    pi2 = 0.6
    pi3 = 0.21
    pi4 = 5.1
    pi5 = 0.49
    pi6 = 0.362
    pi7 = -4.75
    pi8 = -0.4
    pi9 = 0.03

    # Calculate A and B
    A = 2 - (1 + (zeta / pi0)**pi1)**pi2
    B = ((1 + (zeta / pi3)**pi4)**pi5) - 1
    pi25 = 1 - (1 + (zeta / pi6)**pi7)**pi8 - pi9 * np.exp(-((zeta - 0.1)**2) / 0.005)

    # Constants for water properties
    TcH2O = 647.26  # K
    pcH2O = 22.064  # MPa

    # Reduced temperature
    Theta = (T + 273.15) / TcH2O

    # Calculate fsol and pi
    fsol = A + B * Theta
    pi = pi25 * fsol

    # Calculate vapor pressure of water
    tau = 1 - Theta

    A0 = -7.858230
    A1 = 1.839910
    A2 = -11.781100
    A3 = 22.670500
    A4 = -15.939300
    A5 = 1.775160

    lnpi = (A0 * tau + A1 * tau**1.5 + A2 * tau**3 + A3 * tau**3.5 + 
            A4 * tau**4 + A5 * tau**7.5) / (1 - tau)
    pdH2O = np.exp(lnpi) * pcH2O * 1e6  # Convert to Pa

    # Final vapor pressure over the solution
    pd = pi * pdH2O

    # check if the solution is in liquid state
    t_sol = solubility_temperature(x)
    if T < t_sol:
        print(f"Warning: The temperature {T} is below the solubility temperature {t_sol} at x={x}.")

    return pd

    
# DONE
def saturation_concentration(p, T):
    """
    Calculates the saturation concentration of LiCl in water based on the temperature and pressure.

    Author: Dorian Höffner
    Date: 2024-09-17

    Parameters:
        p (float): Pressure in Pa.
        T (float): Temperature in °C.

    Returns:
        float: Saturation concentration in kg LiCl / kg solution.
    """

    # turn off print statemtnes for fsolve
    stout_old = sys.stdout
    devnull = open(os.devnull, 'w')
    sys.stdout = devnull

    # Calculate the saturation concentration
    x_guess = 0.001
    x = fsolve(lambda x: saturation_pressure(x, T) - p, x_guess)[0]

    # turn on print statements again
    sys.stdout = stout_old

    # check if the solution is in liquid state
    t_sol = solubility_temperature(x)
    if T < t_sol:
        print(f"Warning: The temperature {T} is below the solubility temperature {t_sol} at x={x}.")

    return x

# DONE
def density(x, T):
    """
    Calculates the density of LiCl solution based on concentration and temperature.

    Parameters:
    x (float): Mole fraction of NaOH in the solution. [m_LiCl / (m_h2o+m_LiCl)]
    T (float): Temperature in °C.

    Returns:
    float: Density in kg/m^3.
    ---
    Author: Dorian Höffner 11/2024
    Source: CONDE2009 "Aqueous solutions of lithium and calcium chlorides: property formulations for use in air conditioning equipment design" (2009)
    """
    if x > 0.56:
        raise ValueError("Concentration of LiCl cannot be greater than 0.56 kg/kg. Correlation is not valid for this concentration.")

    # Water Correlation
    def rho_h2o(T):
        B = [1.9937718430, 1.0985211604, -0.5094492996, -1.7619124270, -44.9005480267, -723692.2618632]
        rho_h2o_c = 322.0  # kg/m^3 (critical density of water)
        Tc = 647.1  # K
        Theta = (T+273.15) / Tc
        t = 1-Theta
        rho_h2o = rho_h2o_c * (1 + B[0] * t**(1/3) + B[1] * t**(2/3) + B[2] * t**5/3 + B[3] * t**(16/3) + B[4] * t**(43/3) + B[5] * t**(110/3))
        return rho_h2o
    
    # LiCl-H2O Correlation
    c = [1.0, 0.540966, -0.303792, 0.100791]
    rho = rho_h2o(T) * sum([c[i] * (x/(1-x))**i for i in range(4)])

    # check if the solution is in liquid state
    t_sol = solubility_temperature(x)
    if T < t_sol:
        print(f"Warning: The temperature {T} is below the solubility temperature {t_sol} at x={x}.")

    return rho # kg/m^3


# DONE
def specific_heat_capacity(x, T):
    """
    Calculate the specific heat capacity of a LiCl solution (LiCl-H2O) as a function of temperature and concentration.
    
    ---
    x: float
        Concentration of LiCl in the solution [kg/kg]
    T: float
        Temperature of the solution [°C]
    ---
    returns: float
        Specific heat capacity of the solution [kJ/kgK]
    ---
    Last Change: Dorian Höffner
    Author: O. Buchin 03/2011
    Date: 2024-11-04
    Source: Correlation according to CONDE2009 "Aqueous solutions of lithium and calcium chlorides: property formulations for use in air conditioning equipment design" (2009)
    """

    # Coefficients
    a = np.where(T <= 0, 830.54602, 88.7891)
    b = np.where(T <= 0, -1247.52013, -120.1958)
    c = np.where(T <= 0, -68.60350, -16.9264)
    d = np.where(T <= 0, 491.27650, 52.4654)
    e = np.where(T <= 0, -1.80692, 0.10826)
    f = np.where(T <= 0, -137.51511, 0.46988)

    A = 1.43980
    B = -1.24317
    C = -0.12070
    D = 0.12825
    E = 0.62934
    F = 58.5225
    G = -105.6343
    H = 47.7948

    # Theta - reduced temperature
    Theta = (T + 273.15) / 228 - 1

    # cpH2O calculation
    cpH2O = (a + b * Theta**0.02 + c * Theta**0.04 + 
             d * Theta**0.06 + e * Theta**1.8 + f * Theta**8)

    # f1 calculation
    f1 = np.where(x <= 0.31, 
                  A * x + B * x**2 + C * x**3, 
                  D + E * x)

    # f2 calculation
    f2 = F * Theta**0.02 + G * Theta**0.04 + H * Theta**0.06

    # cp calculation
    cp = cpH2O * (1 - f1 * f2)

    # check if the solution is in liquid state
    t_sol = solubility_temperature(x)
    if T < t_sol:
        print(f"Warning: The temperature {T} is below the solubility temperature {t_sol} at x={x}.")

    return cp

# DONE
def dynamic_viscosity(x, T):
    """
    Computes the dynamic viscosity of LiCl with temperature T and concentration x.
    ---
    Parameters:
        T (numpy array): Temperature of the solution (in Celsius) (column vector)
        x (numpy array): Concentration of the solution (0...1) kgSalt/kgSolution (column vector)
    ---
    Returns:
        numpy array: Dynamic viscosity (mPas) (column vector)
    ---
    Last Change: Dorian Höffner 10/2024
    Author: O. Buchin 09/2010
    Source: correlation according to CONDE2009 "Aqueous solutions of lithium and calcium chlorides: property formulations for use in air conditioning equipment design" (2009)
    """

    # Pressure in bar
    p = 1.01325  # bar

    # Coefficients
    eta1 = 0.090481
    eta2 = 1.390262
    eta3 = 0.675875
    eta4 = -0.583517

    # Mass fraction
    xi = x
    zeta = xi / ((1 - xi)**(1 / 0.6))

    # Constants for eta_H2O
    A = 1.0261862
    B = 12481.702
    C = -19510.923
    D = 7065.286
    E = -395.561
    F = 143922.996

    # Theta - reduced temperature
    Theta = (T + 273.15) / 228 - 1
    TcH2O = 647.26  # K
    
    # Placeholder for eta_H2O computation
    etaH2O_0 = 0.001  # This should be replaced with an accurate function for water viscosity

    # Calculate eta_H2O
    eta_H2O = np.where(
        T <= 0,
        etaH2O_0 * (A + B * Theta**0.02 + C * Theta**0.04 + D * Theta**0.08 +
                    E * Theta**2.85 + F * Theta**8),
        etaH2O_0 * np.ones_like(T)  # Replace with actual function for T > 0 if available
    )

    # Adjusted Theta for eta calculation
    Theta_adjusted = (T + 273.15) / TcH2O

    # Calculate dynamic viscosity
    eta = eta_H2O * np.exp(
        eta1 * zeta**3.6 +
        eta2 * zeta +
        eta3 * zeta / Theta_adjusted +
        eta4 * zeta**2
    )

    # check if the solution is in liquid state
    t_sol = solubility_temperature(x)
    if T < t_sol:
        print(f"Warning: The temperature {T} is below the solubility temperature {t_sol} at x={x}.")

    return eta * 1000  # Return result in mPas

# DONE
def diffusion_coefficient(x, T):
    """
    Computes the self diffusion coefficient of a XXXXXX solution.

    Parameters:
        T (np.ndarray or float): Temperature of the solution (in Celsius) (column vector).
        x (np.ndarray or float): Concentration of solution (0 to 1) kgSalt/kgSolution (column vector).
    
    Returns:
        float or np.ndarray: Self diffusion coefficient (m^2/s).

    ---
    Author: O. Buchin 03/2011 (original matlab)
    Last change: Dorian Höffner 11/2024
    Source: correlation according to CONDE2009 "Aqueous solutions of lithium and calcium chlorides: property formulations for use in air conditioning equipment design" (2009)
    """

    # Self Diffusion coefficient for water
    D0 = 1.635e-8  # m^2/s
    TS = 215.05  # K
    gamma = 2.063

    Dw = D0 * (((T + 273.15) / TS) - 1)**gamma

    # Coefficients for the solution
    d1 = 0.52
    d2 = -4.92
    d3 = -0.56

    # Self Diffusion coefficient for solution
    D = Dw * (1 - (1 + (np.sqrt(x) / d1)**d2)**d3)

    # check if the solution is in liquid state
    t_sol = solubility_temperature(x)
    if T < t_sol:
        print(f"Warning: The temperature {T} is below the solubility temperature {t_sol} at x={x}.")

    return D

# TODO
def thermal_conductivity(x, T, p):
    """
    Calculate the thermal conductivity of a NaOH solution as a function of temperature and concentration.
    The function is based on the following publication:
    ---
    x: float
        Concentration of NaOH in the solution [kg/kg]
    T: float
        Temperature of the solution [°C]
    ---
    returns: float
        Thermal conductivity of the solution [W/mK]
    ---
    """
    print("Thermal conductivity function not available.")

    #return lambda_NaOH # [W/mK]

# DONE
def pTDiagram(log=True, invT=True, editablePlot=False, show_percentages=True):
    """
    Plots the pressure-temperature diagram for LiCl-H2O solutions.
    Author: Dorian Höffner 2024-04-26
    Last Change: 2024-09-17
    
    Parameters:
        log (bool): If True, the y-axis will be logarithmic.
        invT (bool): If True, the x-axis will be scaled as -1/T.
        editablePlot (bool): If True, the plot will be editable.
        show_percentages (bool): If True, the concentrations will be labeled.

    Returns:
        None
    """

    # suppress print statements (warnings
    stout_old = sys.stdout
    devnull = open(os.devnull, 'w')
    sys.stdout = devnull
    
    steamTable = XSteam(XSteam.UNIT_SYSTEM_MKS)  # m/kg/sec/°C/bar/W

    # get data to plot crystallization curve
    cryst_data = crystallization_curve(return_data=True)
    cryst_T = np.array([T for x, T in cryst_data]).flatten()
    cryst_p = np.array([saturation_pressure(x, T) for x, T in cryst_data]).flatten()
    
    # Temperature range
    temperaturesC = np.arange(0, 110, 1)
    temperaturesK = temperaturesC + 273.15
    concentrations = np.arange(0.1, 0.59, 0.01)
    cr_temperaturesC = cryst_T
    cr_temperaturesK = cryst_T + 273.15
    
    # Prepare the plot
    plt.figure(dpi=300)
    
    # these temperatures are used for the x-axis
    plotTemperatures = np.arange(0, 101, 10) + 273.15
    
    # Calculate water vapor pressure using XSteam
    waterPressure = [steamTable.psat_t(T - 273.15) * 1e5 for T in temperaturesK]  # convert bar to Pa
    
    # Plot the data
    for x in concentrations:

        # Calculate saturation pressure for each temperature at the given concentration
        p = []
        for T in temperaturesC:
            p.append(saturation_pressure(x, T))  # Assuming PressureNaOH is defined elsewhere

    
        # Set color and line width
        color = "black" if int(np.round(x * 100)) % 10 == 0 else "grey"
        lw = 1.0 if color == "black" else 0.25
        
        # Plotting based on conditions
        temp_plot       = -1/temperaturesK     if invT else temperaturesK
        temp_plot_cryst = -1/(cryst_T+273.15)  if invT else cryst_T+273.15
        ylabel = 'Saturation Pressure [Pa]' if log else 'Saturation Pressure [Pa]'
        xlabel = 'Temperature [°C] (scaled -1/T)' if invT else 'Temperatur [°C]'
        
        # Plot isosteres
        if log:
            plt.semilogy(temp_plot, p, color=color, lw=lw)
        else:
            plt.plot(temp_plot, p, color=color, lw=lw)
        
        plt.ylabel(ylabel)
        plt.xlabel(xlabel)
        plt.xticks(-1/plotTemperatures if invT else plotTemperatures, plotTemperatures - 273.15)
        
        # Label concentrations
        if show_percentages and int(np.round(x * 100)) % 10 == 0 and x>0.19:
            label_pos = temp_plot[-1] + 1e-5 if invT else temp_plot[-1]+2
            plt.text(label_pos, p[-1], f'{x * 100:.0f} %', fontsize=8, color='black')

    # Plotting water line
    if log:
        plt.semilogy(temp_plot, waterPressure, color="grey", linestyle='--',
                     label='Pure Water')
    else:
        plt.plot(temp_plot, waterPressure, color="grey", linestyle='--', label='Pure Water')

    # Plotting crystallization curve
    if log:
        plt.semilogy(temp_plot_cryst, cryst_p, color="gray", linestyle='-',
                    lw=1.0, label='Crystallization Curve', zorder=101)
        # fill area between crystallization curve and the minimum positive value (e.g., 1)
        plt.fill_between(temp_plot_cryst, cryst_p, 1, where=(cryst_p > 1), color='white', zorder=100)
    else:
        plt.plot(temp_plot_cryst, cryst_p, color="gray", linestyle='-',
                 lw=1.0, label='Crystallization Curve', zorder=101)
        # fill area between crystallization curve and a fixed value (e.g., 50)
        plt.fill_between(temp_plot_cryst, cryst_p, 1, where=(cryst_p > 1),  color='white', zorder=100)

    

    # Setting axis limits
    if invT:
        plt.xlim(-1/temperaturesK[0], -1/temperaturesK[-1])
    else:
        plt.xlim(temperaturesK[0], temperaturesK[-1])
    
    if log:
        plt.ylim(220, 1.1e5)
    else:
        plt.ylim(0, max(waterPressure) * 1.1)  # Adjust as needed to make sure all data is visible


    plt.legend()
    
    if not editablePlot:
        plt.show()

    # turn on print statements again
    sys.stdout = stout_old

# DONE
def solubility_temperature(xs):
    """
    Computes the solubility boundary temperature with salt concentration xs.
    ---
    Parameters:
        xs (float or array-like): Concentration of solution (0...1) kgSalt/kgSolution.
    ---
    Returns:
        ts (float or array-like): Temperature of crystallization.
    ---
    Author: O. Buchin 03/2011
    Last change: Dorian Höffner 11/2024
    Source: correlation according to CONDE2009 "Aqueous solutions of lithium and calcium chlorides: property formulations for use in air conditioning equipment design" (2009)
    """

    if np.isscalar(xs):
        xs = np.array([xs])
    

    # Coefficients
    A0 = np.array([0.422088, -0.005340, -0.560360, -0.315220, -1.312310, -1.356800])
    A1 = np.array([-0.090410, 2.015890, 4.723080, 2.882480, 6.177670, 3.448540])
    A2 = np.array([-2.936350, -3.114590, -5.811050, -2.624330, -5.034790, 0.000000])

    # Constants
    TcH2O = 647.26  # K

    # Initialize temperature array
    t = np.zeros((xs.shape[0], 6))

    # Compute temperature for the ice line (k=1)
    k = 0
    theta = A0[k] + A1[k] * xs + A2[k] * xs**2.5
    t[:, k] = theta * TcH2O - 273.15

    # Compute temperatures for hydrates (k=2 to 6)
    for k in range(1, 6):
        theta = A0[k] + A1[k] * xs + A2[k] * xs**2.0
        t[:, k] = theta * TcH2O - 273.15

    # Find the maximum temperature across all hydrate cases
    ts = np.max(t, axis=1)

    if len(ts) == 1:
        ts = ts[0]

    return ts

# DONE
def crystallization_curve(return_data=False):
    """    
    Author: Dorian Höffner
    Date: 2024-11-04

    Parameters:
        return_data (bool): If True, the data will be returned as a list of lists.

    Returns:
        None or list of lists: If return_data is True, the data will be returned as a list of lists (x, T).
    """

    # cryst_data = x, T // create cryst_data based on "solubility_temperate"
    xs_array = np.linspace(0.01, 0.6, 100)
    cryst_data = [[xs, solubility_temperature(xs)] for xs in xs_array]
    temperatures = [item[1] for item in cryst_data]
    
        
    if return_data:
        return cryst_data
    
    else:
        cryst_data = [[xs_array[i]*100, temperatures[i]] for i,_ in enumerate(xs_array)]
        # prepare plot
        plt.figure(figsize=(6,4), dpi=300)
        plt.plot([x[0] for x in cryst_data],
                    [x[1] for x in cryst_data],
                    label='Crystallization Curve',
                    color="black")
        plt.xlabel(r'$\mathrm{CaCl_2}$ Concentration [%]')
        plt.ylabel('Temperature [°C]')
        plt.xlim(0, 60)
        plt.ylim(-100, 200)
        # make beautiful grid
        plt.grid(True)
        plt.minorticks_on()
        plt.grid(which='major', linestyle='-', linewidth='0.2', color='black')
        plt.grid(which='minor', linestyle=':', linewidth='0.1', color='black')
        # add legend (top left)
        plt.legend(loc='upper left')


import numpy as np
from scipy.interpolate import griddata
from scipy.optimize import fsolve
import matplotlib.pyplot as plt


from CoolProp.CoolProp import PropsSI
from pyXSteam.XSteam import XSteam
steamTable = XSteam(XSteam.UNIT_SYSTEM_MKS)  # m/kg/sec/°C/bar/W

import os
import sys

# Contains Parameters for LiBr-H2O Correlations from Patek/Klomfar 2006
class Params_PK:
    def __init__(self):
        # Critical point of pure water
        self.TCritW = 647.096         # K
        self.pCritW = 22.064e6        # Pa
        self.densCritW = 322          # kg/m³
        self.densCritWmol = 17873.727 # mol/m³
        self.cpCritWmol = 76.0226     # J/(mol-K)
        self.enthalpyCritWmol = 37548.5  # J/mol
        self.entropyCritWmol = 79.3933   # J/(mol-K)

        # Triple point of pure water
        self.TTripW = 273.16          # K
        self.pTripW = 611.657         # Pa
        self.densTripW = 999.789      # kg/m³

        # Molar masses
        self.M_W = 0.018015268        # kg/mol
        self.M_LiBr = 0.08685         # kg/mol

        # Regression coefficients

        # Table 4: Pressure
        self.mTab4 = [3, 4, 4, 8, 1, 1, 4, 6]
        self.nTab4 = [0, 5, 6, 3, 0, 2, 6, 0]
        self.tTab4 = [0, 0, 0, 0, 1, 1, 1, 1]
        self.aTab4 = [-241.303, 19175000, -175521000, 32543000, 392.571, -2126.26, 185127000, 1912.16]

        # Table 5: Density
        self.mTab5 = [1, 1]
        self.tTab5 = [0, 6]
        self.aTab5 = [1.746, 4.709]

        # Table 6: Specific heat capacity (cp)
        self.mTab6 = [2, 3, 3, 3, 3, 2, 1, 1]
        self.nTab6 = [0, 0, 1, 2, 3, 0, 3, 2]
        self.tTab6 = [0, 0, 0, 0, 0, 2, 3, 4]
        self.aTab6 = [-14.2094, 40.4943, 111.135, 229.98, 1345.26, -0.014101, 0.0124977, -0.000683209]

        # Table 7: Enthalpy
        self.mTab7 = [1, 1, 2, 3, 6, 1, 3, 5, 4, 5, 5, 6, 6, 1, 2, 2, 2, 5, 6, 7, 1, 1, 2, 2, 2, 3, 1, 1, 1, 1]
        self.nTab7 = [0, 1, 6, 6, 2, 0, 0, 4, 0, 4, 5, 5, 6, 0, 3, 5, 7, 0, 3, 1, 0, 4, 2, 6, 7, 0, 0, 1, 2, 3]
        self.tTab7 = [0, 0, 0, 0, 0, 1, 1, 1, 2, 2, 2, 2, 2, 3, 3, 3, 3, 3, 3, 3, 4, 4, 4, 4, 4, 4, 5, 5, 5, 5]
        self.aTab7 = [
            2.27431, -7.99511, 385.239, -16394, -422.562, 0.113314, -8.33474, -17383.3,
            6.49763, 3245.52, -13464.3, 39932.2, -258877, -0.00193046, 2.80616, -40.4479,
            145.342, -2.74873, -449.743, -12.1794, -0.00583739, 0.23391, 0.341888, 8.85259,
            -17.8731, 0.0735179, -0.00017943, 0.00184261, -0.00624282, 0.00684765
        ]

        # Table 8: Entropy
        self.mTab8 = [1, 1, 2, 3, 6, 1, 3, 5, 1, 2, 2, 4, 5, 5, 6, 6, 1, 3, 5, 7, 1, 1, 1, 2, 3, 1, 1, 1, 1]
        self.nTab8 = [0, 1, 6, 6, 2, 0, 0, 4, 0, 0, 4, 0, 4, 5, 2, 5, 0, 4, 0, 1, 0, 2, 4, 7, 1, 0, 1, 2, 3]
        self.tTab8 = [0, 0, 0, 0, 0, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 2, 3, 3, 3, 3, 4, 4, 4, 4, 4, 5, 5, 5, 5]
        self.aTab8 = [
            1.53091, -4.52564, 698.302, -21666.4, -1475.33, 0.0847012, -6.59523,
            -29533.1, 0.00956314, -0.188679, 9.31752, 5.78104, 13893.1, -17176.2,
            415.108, -55564.7, -0.00423409, 30.5242, -1.6762, 14.8283, 0.00303055,
            -0.040181, 0.149252, 2.5924, -0.177421, -0.000069965, 0.000605007,
            -0.00165228, 0.00122966
        ]



# LiBr property functions
# DONE
def saturation_temperature(x, p):
    """
    Calculate the boiling point temperature of an aqueous Lithium Bromide solution.
    ---
    Parameters:
        p (float): Pressure in Pa.
        x (float): Concentration in kg LiBr/kg solution, i.e., 0.0 to 0.75.
    ---
    Returns:
        float: Boiling point temperature in °C.
    """

    # check validity of concentration
    if x > 0.75:
        raise ValueError("Concentration out of valid range (0.0 to 0.75).")

    # use saturation_pressure and fsolve to find the temperature
    def f(T):
        return saturation_pressure(x, T) - p
    
    # initial guess for fsolve
    T_guess = 50 # [°C]

    # find the saturation temperature
    T = fsolve(f, T_guess)[0]

    return T # [°C]

# def enthalpy(x,T):
#     """
#     Calculate the enthalpy of a LiBr solution at a given temperature and concentration.
#     ---
#     Parameters:
#         x (float): Concentration of LiBr in the solution (kg LiBr/kg solution), should be between 0.00 and 0.75.
#         T (float): Temperature in Kelvin, should be between 273.15 K and 500 K.
#     ---
#     Returns:
#         float: Enthalpy in kJ/kg
#     ---
#     """
#     tempK = T + 273.15  # convert to Kelvin
    
#     #Specific enthalpy of a lithium bromide solution at 300 K and 1 atm
#     return PropsSI('H','T',tempK,'P',101325,f'INCOMP::LiBr-{x*100}%')



# DONE (40%-75%)
def enthalpy(konz, tempC):
    """
    Calculate the enthalpy of an H2O-LiBr solution at a given temperature and concentration.
    ---
    Parameters:
        konz (float): Salt mass fraction in kg LiBr/kg solution (0.4 <= konz <= 0.75).
        tempC (float): Temperature in degrees Celsius (°C).
    ---
    Returns:
        float: Enthalpy in kJ/kg. Returns -1 if the temperature is out of range (0°C to 190°C).
            Returns -2 if the concentration is out of range (0.4 to 0.75).
    ---
    Notes:
        This function calculates the enthalpy based on the method described in:
        Feuerecker, Günther: Entropieanalyse für Wärmepumpensysteme: Methoden und Stoffdaten.
        Institut für Festkörperphysik und Technische Physik der Technischen Universität München,
        München 1994, S. 131 Gleichung (9.13).
    ---
    History:
    - Originally programmed for MATLAB by Jan Albers, TU Berlin, jan.albers@tu-berlin.de, on 2003-12-03.
    - standardized API, added ValueErrors by Dorian Höffner 11/2024
    """

    # Coefficients for the enthalpy calculation
    coefhli = [[-954.8,    47.7739,    -1.59235,    2.09422e-2, -7.689e-5],
               [-0.3293,    4.076e-2,  -1.36e-5,   -7.1366e-6],
               [7.4285e-3, -1.5144e-4,  1.3555e-6],
               [-2.269e-6]]

    temp = tempC + 273.15  # convert to Kelvin

    if temp < 273.15 or temp > 190 + 273.15:
        raise ValueError("Temperature out of valid range (0°C to 190°C).")
    if konz < 0.4 - 10 ** (-9) or konz > 0.75 + 10 ** (-9):
        raise ValueError("Concentration out of valid range (0.0 to 0.75).")

    # Vorbereiten der Stoffparameter,
    konz = konz * 100  # G.F. rechnet in # nicht in kg/kg, daher Konvertierung

    a = 0
    for k in range(len(coefhli[0])):
        a += konz ** k * coefhli[0][k]

    b = 0
    for k in range(len(coefhli[1])):
        b += konz ** k * coefhli[1][k]
    c = 0
    for k in range(len(coefhli[2])):
        c += konz ** k * coefhli[2][k]

    return a + temp * b + temp ** 2 * c + coefhli[3][0] * temp ** 3
    

    

    return h # enthalpy in [kJ/kg]

# TODO
def differential_enthalpy_AD(x, T):
    """
    Calculates the differential enthalpy [kJ/kgH2O] of a CaCl2 solution based on concentration and temperature.
    How much additional energy (in comparison to pure vaporization enthalpy) is needed to vaporize a certain amount of water in a CaCl2 solution.
    
    Parameters:
        x (float): Concentration of CaCl2 in the solution [kg/kg].
        T (float): Temperature of the solution [°C].

    Returns:
        float: Differential enthalpy [kJ/kgH2O].

    Author: Dorian Höffner 11/2024
    Source: 
    """

    return

# DONE
def saturation_pressure(x, T):
    """
    Calculate the equilibrium pressure of an H2O-LiBr solution.
    Parameters:
        x (float): Mass fraction of LiBr in the solution (kg LiBr/kg solution), should be between 0.00 and 0.75.
        T (float): Temperature in Kelvin, should be between 273.15 K and 500 K.
    Returns:
        float: Equilibrium pressure in bar. Returns -1 if temperature is out of valid range, 
            and -2 if mass fraction is out of valid range.
    References:
        J. Patek and J. Klomfar, "A computationally effective formulation of the thermodynamic properties 
        of LiBr-H2O solutions from 273 to 500 K over full composition range," to be published in IJR.
    History:
        - Programmed by Jan Albers on DD.MM.2010
        - Validity limits added on 02.08.2012 by J.A.
        - standardized API and added ValueError by Dorian Höffner 2024-11-05
    Restrictions:
        - The function is only valid for temperatures between 273.15 K and 500 K.
        - The function is only valid for mass fractions between 0.00 and 0.75.
    """

    def psatW_local(temp):
        """
        Calculates water vapor pressure based on matlab package h2o_IAPWS97.
        Author: Dorian Höffner 2024-11-21
        """
        
        # Coefficients
        nreg4_9  = -0.23855557567849
        nreg4_10 = 650.17534844798
        nreg4_1  = 1167.0521452767
        nreg4_2  = -724213.16703206
        nreg4_3  = -17.073846940092
        nreg4_4  = 12020.82470247
        nreg4_5  = -3232555.0322333
        nreg4_6  = 14.91510861353
        nreg4_7  = -4823.2657361591
        nreg4_8  = 405113.40542057

        dl = temp + nreg4_9 / (temp - nreg4_10)
        Aco = dl ** 2 + nreg4_1 * dl + nreg4_2
        Bco = nreg4_3 * dl ** 2 + nreg4_4 * dl + nreg4_5
        cco = nreg4_6 * dl ** 2 + nreg4_7 * dl + nreg4_8
        pSatW = (2 * cco / (-Bco + (Bco ** 2 - 4 * Aco * cco) ** 0.5)) ** 4 * 10
    
        return pSatW
    
    T = T + 273.15  # convert to Kelvin

    # get necessary parameters from Params class (Patek/Klomfar 2006)
    params  = Params_PK()
    M_LiBr  = params.M_LiBr
    M_W     = params.M_W
    TCritW  = params.TCritW
    aTab4   = params.aTab4
    mTab4   = params.mTab4
    nTab4   = params.nTab4
    tTab4   = params.tTab4

    # Validity Check
    if T < 273.15 or T > 500:
        raise ValueError(f"Temperature (T={T}) out of valid range (273.15 K to 500 K).")
    if x < 0.00 or x > 0.75:
        raise ValueError(f"Mass fraction (x={x}) out of valid range (0.00 to 0.75).")

    # Calculation
    tsum = 0
    
    x_mol = (x/M_LiBr)/(x/M_LiBr + (1-x)/M_W)
    
    for k in range(len(aTab4)):
        tsum += aTab4[k] * x_mol ** mTab4[k] * (0.40 - x_mol) ** nTab4[k] * (T / TCritW) ** tTab4[k]

    # return psat * 1e5  # convert bar to Pa
    psat = psatW_local(T - tsum)  # in bar

    return psat * 1e5  # convert bar to Pa



# DONE
def saturation_concentration(p, T):
    """
    Calculates the saturation concentration of LiBr in water based on the temperature and pressure.
    Source: Feuerecker et al. 1994
    
    Author: Dorian Höffner
    Date: 2024-09-17
    ---
    Parameters:
        p (float): Pressure in Pa.
        T (float): Temperature in °C.
    ---
    Returns:
        float: Saturation concentration in kg LiBr / kg solution.
    ---
    Restrictions:
        - The function is only valid for temperatures between 273.15 K and 500 K.
    """

    if T < 0 or T > 227.85:
        raise ValueError("Temperature out of valid range (0°C to 227.85°C).")

    # Calculate the saturation concentration
    x = 0.001
    while saturation_pressure(x, T) - p > 1:   # 1 Pa tolerance
        x += 0.001

    if x > 0.75:
        print(f"Warning: Saturation concentration is above 75% at {T}°C and {p} Pa. This is outside the publication's validity range.")

    return x

# DONE
def density(konz, tempC):
    """
    Calculate the density of a water-LiBr solution.
    This function calculates the density of a water-LiBr (Lithium Bromide) solution 
    based on the given mass fraction and temperature using the correlations provided 
    by Günther Feuerecker and corrections from private communication between F. Ziegler 
    and G. Feuerecker.
    ---
    Parameters:
        konz (float): Mass fraction of LiBr in the solution (kg_LiBr/kg_Sol). Valid range is 0.0 to 0.75.
        tempC (float): Temperature in degrees Celsius. Valid range is 0°C to 190°C.
    ---
    Returns:
        float: Density of the water-LiBr solution in kg/m³.
    ---
    Raises:
        ValueError: If the temperature is out of the valid range (0°C to 190°C).
        ValueError: If the mass fraction is out of the valid range (0.0 to 0.75).
    ---
    Notes:
    - If the mass fraction is below 0.4, the results are extrapolated.
    - The density of pure water is calculated using coefficients from Landolt-Börnstein.
    - The function includes corrections from private communication between F. Ziegler and G. Feuerecker (1995).
    References:
    - Feuerecker, Günther: Entropieanalyse für Wärmepumpensysteme: Methoden und Stoffdaten. 
      Institut für Festkörperphysik und Technische Physik der Technischen Universität München, 
      München 1994.
    - Landolt-Börnstein: II. Band Eigenschaften der Materie in ihren Aggregatzuständen.
      1. Teil Mechanisch-thermische Zustandsgrößen. 6. Auflage Berlin 1971.
    """
    # Historie
    # J.A. Nov. 2003 programmiert
    # J.A. Feb. 2004 Begrenzung der Wasserdichte (H2O) auf Siededichte eingefügt
    # J.A. Apr. 2012 Correction from private communication between F. Ziegler and G. Feuerecker 1995 included
    #                Implementation of pure water coefficients from Landolt-Börnstein (as used by Peter Müller)
    # Dorian Höffner: standardized API, added ValueError, added extrapolation for mass fraction below 40%
    #
    # Conversion from input temperature in Kelvin to Celsius used in the correlations of G.F.
    temp = tempC + 273.15

    if tempC < 0 or tempC > 190:
        raise ValueError("Temperature out of valid range (0°C to 190°C).")
    if konz < 0.4 - 10 ** (-9):
        print("Warning: Results for density were extrapolated for mass fractions from 0.0 to 0.4.")
    if konz > 0.75 + 10 ** (-9):
        raise ValueError("Mass fraction out of valid range (0.4 to 0.75).")
    if konz < 0.0:
        raise ValueError("Mass fraction out of valid range (0.0 to 0.75).")

    # New from Apr 2012 on: The same water density is used here as was done by Peter Müller
    # to derive the coefficients for the density of H_2O/LiBr-solution.
    #
    # Peter Müller used the coeffients of
    # Landolt-Börnstein: II. Band Eigenschaften der Materie in ihren Aggregatzuständen.
    # 1. Teil Mechanisch-thermische Zustandsgrößen. 6. Auflage Berlin 1971
    a_1 = 3.9863  # K
    a_2 = 508929.2  # K**2
    a_3 = 288.9414  # K
    a_4 = 68.12963  # K
    a_5 = 0.999973  # g/cm**3

    dens_H2O = (1 - (tempC - a_1) ** 2 / a_2 * (tempC + a_3) / (tempC + a_4)) * a_5
    # todo use denseH2O  function instead?

    # published equation of G.F.
    # densH2OLiBr_GF = dens_H2O  / 2  * (exp(1.2  *  konz) + exp((0.842+1.6414e-3 * tempC) * konz))

    # including the correction from private communication between F. Ziegler and G. Feuerecker 1995
    # (the power of 2 was missing at last konz value)
    return (dens_H2O / 2.0 * (np.exp(1.2 * konz) + np.exp((0.842 + 1.6414e-3 * tempC) * konz ** 2))) * 1000

# # DONE
# def density OLD NOT WORKING PROPERLY(x, T):
#     """
#     Calculate the density of H2O-LiBr solution.
#     This function computes the density of a H2O-LiBr solution based on the 
#     temperature and mass fraction of LiBr in the solution.
#     ---
#     Parameters:
#         x (float): Mass fraction of LiBr in the solution (kg_LiBr/kg_Solution), e.g., 0 to 0.75.
#         T (float): Temperature in °C.
#     ---
#     Returns:
#         float: Density of the H2O-LiBr solution in kg/m³.
#     ---
#     References:
#         Chua et al. 1999 "Improved thermodynamic property Æelds of LiBr±H2O solution"
#     ---
#     Author: Dorian Höffner 2024-11-05
#     """

#     T = T + 273.15  # convert to Kelvin
    
#     G = [[9-99100e2,      7.74931e0, 5.36509e-3, 1.34988e-3, -3.08671e-6],
#          [-2.39865e-2,  -1.28346e-2, 2.07232e-4, -9.08213e-6, 9.94788e-8],
#          [-3.90453e-3,  -5.55855e-5, 1.09879e-5, -2.39834e-7, 1.53514e-9]]
#     G = np.array(G)

#     delta_rho = np.sum([x**(i) * (G[0,i] + T * G[1,i] + T**2 * G[2,i]) for i in range(5)])
#     rho = delta_rho + steamTable.rhoL_t(T-273.15)  # kg/m³

#     return rho

    



# TODO
def specific_heat_capacity(x, T):
    """
    Calculate the specific heat capacity of a CaCl2 solution (CaCl2-H2O) as a function of temperature and concentration.
    
    ---
    x: float
        Concentration of CaCl2 in the solution [kg/kg]
    T: float
        Temperature of the solution [°C]
    ---
    returns: float
        Specific heat capacity of the solution [kJ/kgK]
    ---
    Author: Dorian Höffner
    Date: 2024-11-04
    Source: Conde 2009 "Aqueous solutions of lithium and calcium chlorides: property formulations for use in air conditioning equipment design" (2009)
    """


# TODO
def dynamic_viscosity(x, T):
    """
    Calculate the dynamic viscosity of a LiBr solution as a function of temperature and concentration.
    The function is based on the following publication: "Aqueous solutions of lithium and calcium chlorides: property formulations for use in air conditioning equipment design" (2009)
    ---
    x: float
        Concentration of LiBr in the solution [kg/kg]
    T: float
        Temperature of the solution [°C]
    ---
    returns: float
        Dynamic viscosity of the solution [Pa s]
    ---
    Last Change: Dorian Höffner 10/2024
    Author: O. Buchin 09/2010
    """


# TODO
def diffusion_coefficient(x, T):
    """
    Computes the self diffusion coefficient of a CaCl2 solution.

    Parameters:
        T (np.ndarray or float): Temperature of the solution (in Celsius) (column vector).
        x (np.ndarray or float): Concentration of solution (0 to 1) kgSalt/kgSolution (column vector).
    
    Returns:
        float or np.ndarray: Self diffusion coefficient (m^2/s).

    ---
    Author: 
    Last change: Dorian Höffner 11/2024
    Source: 
    """
    

# TODO
def thermal_conductivity(x, T, p):
    """
    Calculate the thermal conductivity of a LiBr solution as a function of temperature and concentration.
    The function is based on the following publication:
    ---
    x: float
        Concentration of LiBr in the solution [kg/kg]
    T: float
        Temperature of the solution [°C]
    ---
    returns: float
        Thermal conductivity of the solution [W/mK]
    ---
    """


# TODO
def pTDiagram(log=True, invT=True, editablePlot=False, show_percentages=True):
    """
    Plots the pressure-temperature diagram for LiBr-H2O solutions.
    Author: Dorian Höffner 2024-04-26
    Last Change: 2024-09-17
    
    Parameters:
        log (bool): If True, the y-axis will be logarithmic.
        invT (bool): If True, the x-axis will be scaled as -1/T.
        editablePlot (bool): If True, the plot will be editable.
        show_percentages (bool): If True, the concentrations will be labeled.

    Returns:
        None
    """

    # turn off print statements
    devnull = open(os.devnull, 'w')
    stdout_old = sys.stdout
    sys.stdout = devnull
    
    
    steamTable = XSteam(XSteam.UNIT_SYSTEM_MKS)  # m/kg/sec/°C/bar/W

    # get data to plot crystallization curve
    cryst_data = crystallization_curve(return_data=True)
    cryst_T = np.array([T for x, T in cryst_data]).flatten()
    cryst_p = np.array([saturation_pressure(x, T) for x, T in cryst_data]).flatten()
    
    # Temperature range
    temperaturesC = np.arange(0, 110, 1)
    temperaturesK = temperaturesC + 273.15
    concentrations = np.arange(0.1, 0.75, 0.01)
    cr_temperaturesC = cryst_T
    cr_temperaturesK = cryst_T + 273.15
    
    # Prepare the plot
    plt.figure(dpi=300)
    
    # these temperatures are used for the x-axis
    plotTemperatures = np.arange(0, 101, 10) + 273.15
    
    # Calculate water vapor pressure using XSteam
    waterPressure = [steamTable.psat_t(T - 273.15) * 1e5 if T>273.15 else np.nan for T in temperaturesK]  # convert bar to Pa

    
    # Plot the data
    for x in concentrations:

        # Calculate saturation pressure for each temperature at the given concentration
        p = []
        for T in temperaturesC:
            p.append(saturation_pressure(x, T))

    
        # Set color and line width
        color = "black" if int(np.round(x * 100)) % 10 == 0 else "grey"
        lw = 1.0 if color == "black" else 0.25
        
        # Plotting based on conditions
        temp_plot       = -1/temperaturesK     if invT else temperaturesK
        temp_plot_cryst = -1/(cryst_T+273.15)  if invT else cryst_T+273.15
        ylabel = 'Saturation Pressure [Pa]' if log else 'Saturation Pressure [Pa]'
        xlabel = 'Temperature [°C] (scaled -1/T)' if invT else 'Temperatur [°C]'
        
        # Plot isosteres
        if log:
            plt.semilogy(temp_plot, p, color=color, lw=lw)
        else:
            plt.plot(temp_plot, p, color=color, lw=lw)
        
        plt.ylabel(ylabel)
        plt.xlabel(xlabel)
        plt.xticks(-1/plotTemperatures if invT else plotTemperatures, plotTemperatures - 273.15)
        
        # Label concentrations
        if show_percentages and int(np.round(x * 100)) % 10 == 0 and x>0.29:
            label_pos = temp_plot[-1] + 1e-5 if invT else temp_plot[-1]+2
            plt.text(label_pos, p[-1], f'{x * 100:.0f} %', fontsize=8, color='black')

    # Plotting water line
    if log:
        plt.semilogy(temp_plot, waterPressure, color="grey", linestyle='--',
                     label='Pure Water')
    else:
        plt.plot(temp_plot, waterPressure, color="grey", linestyle='--', label='Pure Water')

    # Plotting crystallization curve
    if log:
        plt.semilogy(temp_plot_cryst, cryst_p, color="gray", linestyle='-',
                    lw=1.0, label='Crystallization Curve', zorder=101)
        # fill area between crystallization curve and the minimum positive value (e.g., 1)
        plt.fill_between(temp_plot_cryst, cryst_p, 1, where=(cryst_p > 1), color='white', zorder=100)
    else:
        plt.plot(temp_plot_cryst, cryst_p, color="gray", linestyle='-',
                 lw=1.0, label='Crystallization Curve', zorder=101)
        # fill area between crystallization curve and a fixed value (e.g., 50)
        plt.fill_between(temp_plot_cryst, cryst_p, 1, where=(cryst_p > 1),  color='white', zorder=100)

    

    # Setting axis limits
    if invT:
        plt.xlim(-1/temperaturesK[0], -1/temperaturesK[-1])
    else:
        plt.xlim(temperaturesK[0], temperaturesK[-1])
    
    if log:
        plt.ylim(100, 1.4e5)
    else:
        plt.ylim(0, max(waterPressure) * 1.1)  # Adjust as needed to make sure all data is visible


    plt.legend()
    
    if not editablePlot:
        plt.show()

    # turn on print statements again
    sys.stdout = stdout_old

# DONE (56%-75%)
def solubility_temperature(x):
    """
    Calculate the crystallization temperature of LiBr solution in water.
    
    Parameters:
    x (float or np.array): Concentration of LiBr in kg/kg
    
    Returns:
    float or np.array: Crystallization temperature in °C, or -1 if outside valid concentration range.
    
    Valid concentration range:
    0.5681 <= x (kg/kg) <= 0.75
    """

    # catch invalid concentration
    if np.any((x < 0.5681) | (x > 0.75)):
        raise ValueError("Concentration out of valid range (0.5681 to 0.75).")

    # Parameters
    mu_1 = 0.660036363636364
    mu_2 = 0.0521377438043144
    
    # Adjusted concentration
    x_dach = (x - mu_1) / mu_2
    
    # Polynomial terms
    part1 = -1.25354043437046 * x_dach**7 + 1.60535142980859 * x_dach**6 + 9.50132460833796 * x_dach**5 - 10.9718095175445 * x_dach**4
    part2 = -23.0924483393181 * x_dach**3 + 23.9376211870673 * x_dach**2 + 57.4166682907763 * x_dach + 55.0110013350386
    
    # Crystallization temperature in Kelvin
    t_cryst_H2OLiBr_DBGF = part1 + part2 + 273.15
    
    # Validity check
    if np.any((x < 0.5681 - 0.0005) | (x > 0.75 + 0.0005)):
        t_cryst_H2OLiBr_DBGF = -1

    t_solubility = t_cryst_H2OLiBr_DBGF - 273.15
    
    return t_solubility


    
# TODO
def crystallization_curve(return_data=False):
    """    
    Author: Dorian Höffner
    Date:
    Data Source: Data was extracted from plot in publication.
    Publication:

    Parameters:
        return_data (bool): If True, the data will be returned as a list of lists.

    Returns:
        None or list of lists: If return_data is True, the data will be returned as a list of lists (x, T).
    """

    # cryst_data = x, T // create cryst_data based on "solubility_temperate"
    xs_array = np.linspace(0.5681, 0.75, 100)
    cryst_data = [[xs, solubility_temperature(xs)] for xs in xs_array]
    temperatures = [item[1] for item in cryst_data]
    
        
    if return_data:
        return cryst_data
    
    else:
        cryst_data = [[xs_array[i]*100, temperatures[i]] for i,_ in enumerate(xs_array)]
        # prepare plot
        plt.figure(figsize=(6,4), dpi=300)
        plt.plot([x[0] for x in cryst_data],
                    [x[1] for x in cryst_data],
                    label='Crystallization Curve',
                    color="black")
        plt.xlabel(r'$\mathrm{CaCl_2}$ Concentration [%]')
        plt.ylabel('Temperature [°C]')
        plt.xlim(0, 75)
        plt.ylim(-60, 200)
        # make beautiful grid
        plt.grid(True)
        plt.minorticks_on()
        plt.grid(which='major', linestyle='-', linewidth='0.2', color='black')
        plt.grid(which='minor', linestyle=':', linewidth='0.1', color='black')
        # add legend (top left)
        plt.legend(loc='upper left')


from setuptools import setup, find_packages

setup(
    name="absorptionlib",
    version="1.0.2",
    packages=find_packages(),  # Automatically finds the `propertiesNaOH` package
    install_requires=[
        'numpy',
        'matplotlib',
        'pyXSteam',  # For the pyXSteam package
    ],
    author="Dorian Höffner",
    author_email="dorian.hoeffner@tu-berlin.de",
    description="A package for substance properties for absorption systems",
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    #url="https://github.com/dorianhoeffner/propertiesNaOH",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)
